# jhbuild - a tool to ease building collections of source packages
# Copyright (C) 2008  Frederic Peters
#
#   build.py: custom builder pages
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import urllib, time

from twisted.web import html
from twisted.web.util import Redirect

from buildbot.status.web.builder import BuildersResource, StatusResourceBuilder
from buildbot.status.web.base import make_row, make_force_build_form, \
             path_to_slave, path_to_builder

from build import JhBuildsResource

class JhStatusResourceBuilder(StatusResourceBuilder):
    def getTitle(self, request):
        buildbot_service = request.site.buildbot_service
        builder_name = self.builder_status.getName()
        for slave in buildbot_service.slaves:
            if builder_name.endswith(slave):
                slave_name = slave
                module_name = builder_name[:-len(slave)-1]
                break
        else:
            slave_name = None
            module_name = None
        status = self.getStatus(request)
        p = status.getProjectName()
        if slave_name:
            return '%s - %s @ %s' % (
                    p, module_name, slave_name)
        return '%s - %s' % (p, builder_name)

    def body(self, req):
        b = self.builder_status
        control = self.builder_control
        status = self.getStatus(req)

        slaves = b.getSlaves()
        connected_slaves = [s for s in slaves if s.isConnected()]

        projectName = status.getProjectName()

        data = ''

        # the first section shows builds which are currently running, if any.
        current = b.getCurrentBuilds()
        if current:
            data += "<h2>Currently Building:</h2>\n"
            data += "<ul>\n"
            for build in current:
                data += " <li>" + self.build_line(build, req) + "</li>\n"
            data += "</ul>\n"
        else:
            data += "<h2>No current builds</h2>\n"

        # Then a section with the last 5 builds, with the most recent build
        # distinguished from the rest.

        data += "<h2>Recent Builds</h2>\n"
        data += "<ul>\n"
        numbuilds = int(req.args.get('numbuilds', ['5'])[0])
        for i,build in enumerate(b.generateFinishedBuilds(num_builds=int(numbuilds))):
            data += " <li>" + self.make_line(req, build, False) + "</li>\n"
            if i == 0:
                data += "<br />\n" # separator
                # TODO: or empty list?
        data += "</ul>\n"

        data += "<h2>Buildslaves:</h2>\n"
        data += "<ol>\n"
        for slave in slaves:
            slaveurl = path_to_slave(req, slave)
            data += "<li><b><a href=\"%s\">%s</a></b>: " % (html.escape(slaveurl), html.escape(slave.getName()))
            if slave.isConnected():
                data += "CONNECTED\n"
                if slave.getAdmin():
                    data += make_row("Admin:", html.escape(slave.getAdmin()))
                if slave.getHost():
                    data += "<span class='label'>Host info:</span>\n"
                    data += html.PRE(html.escape(slave.getHost()))
            else:
                data += ("NOT CONNECTED\n")
            data += "</li>\n"
        data += "</ol>\n"

        if control is not None and connected_slaves:
            forceURL = path_to_builder(req, b) + '/force'
            data += make_force_build_form(forceURL, self.isUsingUserPasswd(req))
        elif control is not None:
            data += """
            <p>All buildslaves appear to be offline, so it's not possible
            to force this build to execute at this time.</p>
            """

        if control is not None:
            pingURL = path_to_builder(req, b) + '/ping'
            data += """
            <form method="post" action="%s" class='command pingbuilder'>
            <p>To ping the buildslave(s), push the 'Ping' button</p>

            <input type="submit" value="Ping Builder" />
            </form>
            """ % pingURL

        return data


    def getChild(self, path, req):
        if path == 'builds':
            return JhBuildsResource(self.builder_status, self.builder_control)
        return StatusResourceBuilder.getChild(self, path, req)


class JhBuildersResource(BuildersResource):
    def getChild(self, path, req):
        if path == '':
            return Redirect('..')
        s = self.getStatus(req)
        if path in s.getBuilderNames():
            builder_status = s.getBuilder(path)
            builder_control = None
            c = self.getControl(req)
            if c:
                builder_control = c.getBuilder(path)
            return JhStatusResourceBuilder(builder_status, builder_control)
        return BuildersResource.getChild(self, path, req)



# jhbuild - a tool to ease building collections of source packages
# Copyright (C) 2001-2006  James Henstridge
#
#   trayicon.py: simple wrapper for zenity based tray icons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import sys
import os
import subprocess
import signal
try:
    import dbus
except ImportError:
    dbus = None

class TrayIcon:
    proc = None

    def __init__(self, config):
        if dbus is None:
            return
        if config and config.notrayicon:
            return
        if not os.environ.get('DISPLAY'):
            return

        try:
            bus = dbus.SessionBus()
            proxy = bus.get_object('org.freedesktop.Notifications',
                                   '/org/freedesktop/Notifications')
            notify_iface = dbus.Interface(proxy, dbus_interface='org.freedesktop.Notifications')
            caps = notify_iface.GetCapabilities()
            for item in caps:
                if item == "persistence":
                    return
        except:
            pass

        try:
            self._run_zenity()
        except AttributeError:
            # 'cStringIO.StringO' object has no attribute 'fileno'
            pass

    def __del__(self):
        if self.proc:
            self.close()

    def _run_zenity(self):
        # run zenity with stdout and stderr directed to /dev/null
        def preexec():
            null = open('/dev/null', 'w')
            try:
                os.dup2(null.fileno(), sys.stdout.fileno())
                os.dup2(null.fileno(), sys.stderr.fileno())
            finally:
                null.close()
            os.setsid()
        try:
            self.proc = subprocess.Popen(['zenity', '--notification',
                                          '--listen'],
                                         close_fds=True,
                                         preexec_fn=preexec,
                                         stdin=subprocess.PIPE)
        except (OSError, IOError):
            self.proc = None

    def close(self):
        status = None
        if self.proc:
            self.proc.stdin.close()
            try:
                os.kill(self.proc.pid, signal.SIGTERM)
            except OSError:
                pass
            status = self.proc.wait()
            self.proc = None
        return status

    def _send_cmd(self, cmd):
        if not self.proc: return
        try:
            self.proc.stdin.write(cmd)
            self.proc.stdin.flush()
        except (IOError, OSError) as err:
            self.close()
    def set_icon(self, icon):
        self._send_cmd('icon: %s\n' % icon)
    def set_tooltip(self, tooltip):
        self._send_cmd('tooltip: %s\n' % tooltip.encode('utf-8'))
    def set_visible(self, visible):
        if visible:
            visible = 'true'
        else:
            visible = 'false'
        self._send_cmd('visible: %s\n' % visible)



# jhbuild - a tool to ease building collections of source packages
# Copyright (C) 2001-2004  James Henstridge
#
#   autobuild.py: build logic for a non-interactive reporting build
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import os
import time
import subprocess
import sys
import locale
import socket

from jhbuild.utils import cmds
from jhbuild.errors import CommandError
import buildscript

import xmlrpclib
import zlib
from cStringIO import StringIO

from tinderbox import get_distro
from terminal import TerminalBuildScript, trayicon, t_bold, t_reset
import jhbuild.moduleset

def escape(string):
    return string.replace('&','&amp;').replace('<','&lt;').replace('>','&gt;')

def fix_encoding(string):
    charset = locale.getpreferredencoding()
    s = 'VERY BORKEN ENCODING'
    for encoding in [charset, 'utf-8', 'iso-8859-15']:
        try:
            s = unicode(string, encoding)
        except:
            continue
        break
    return s.encode('us-ascii', 'xmlcharrefreplace')

def compress_data(data):
    c_data = zlib.compress(data)
    return xmlrpclib.Binary(c_data)

class ServerProxy(xmlrpclib.ServerProxy):
    verbose_timeout = False

    def __request(self, methodname, params):
        ITERS = 10
        for i in range(ITERS):
            try:
                return xmlrpclib.ServerProxy.__request(self, methodname, params)
            except xmlrpclib.ProtocolError as e:
                if e.errcode != 500:
                    raise
            except socket.error as e:
                pass
            if i < ITERS-1:
                if self.verbose_timeout:
                    print >> sys.stderr, _('Server Error, retrying in %d seconds') % ((i+1)**2)
                time.sleep((i+1)**2)
            else:
                if self.verbose_timeout:
                    print >> sys.stderr, _('Server Error, aborting')
                raise e
            

class AutobuildBuildScript(buildscript.BuildScript, TerminalBuildScript):
    xmlrpc_report_url = None
    verbose = False

    def __init__(self, config, module_list, module_set=None):
        buildscript.BuildScript.__init__(self, config, module_list, module_set=module_set)
        self.xmlrpc_report_url = config.autobuild_report_url
        self.verbose = config.verbose
        self.server = None
        self.modulefp = None
        self.phasefp = None
        self.modules = {}

        # cleanup environment
        os.environ['TERM'] = 'dumb'
        os.environ['LANG'] = 'C'
        for k in os.environ.keys():
            if k.startswith('LC_'):
                os.environ[k] = 'C'

        if self.verbose:
            self.trayicon = trayicon.TrayIcon(config)

    def message(self, msg, module_num=-1, skipfp = False):
        '''Display a message to the user'''
        if not skipfp:
            if self.phasefp:
                fp = self.phasefp
            elif self.modulefp:
                fp = self.modulefp
            else:
                fp = None

            if fp:
                fp.write(msg + '\n')

        if self.verbose:
            TerminalBuildScript.message(self, msg, module_num)

    def set_action(self, action, module, module_num=-1, action_target=None):
        if module_num == -1:
            module_num = self.module_num
        if not action_target:
            action_target = module.name
        self.message('%s %s' % (action, action_target), module_num, skipfp = True)

    def execute(self, command, hint=None, cwd=None, extra_env=None):
        kws = {
            'close_fds': True
            }
        if isinstance(command, (str, unicode)):
            displayed_command = command
            kws['shell'] = True
        else:
            displayed_command = ' '.join(command)

        self.phasefp.write('<span class="command">%s</span>\n' % escape(displayed_command))
        if self.verbose:
            print ' $', displayed_command

        kws['stdin'] = subprocess.PIPE
        kws['stdout'] = subprocess.PIPE
        kws['stderr'] = subprocess.PIPE
        if hint in ('cvs', 'svn', 'hg-update.py'):
            def format_line(line, error_output, fp=self.phasefp):
                if line[-1] == '\n': line = line[:-1]
                if self.verbose:
                    print line
                if line.startswith('C '):
                    fp.write('<span class="conflict">%s</span>\n'
                                        % escape(line))
                else:
                    fp.write('%s\n' % escape(line))
            kws['stderr'] = subprocess.STDOUT
        else:
            def format_line(line, error_output, fp=self.phasefp):
                if line[-1] == '\n': line = line[:-1]
                if self.verbose:
                    if error_output:
                        print >> sys.stderr, line
                    else:
                        print line
                if error_output:
                    fp.write('<span class="error">%s</span>\n'
                                        % escape(line))
                else:
                    fp.write('%s\n' % escape(line))

        if cwd is not None:
            kws['cwd'] = cwd

        if extra_env is not None:
            kws['env'] = os.environ.copy()
            kws['env'].update(extra_env)

        command = self._prepare_execute(command)

        try:
            p = subprocess.Popen(command, **kws)
        except OSError as e:
            self.phasefp.write('<span class="error">' + _('Error: %s') % escape(str(e)) + '</span>\n')
            raise CommandError(str(e))

        cmds.pprint_output(p, format_line)
        if p.returncode != 0:
            raise CommandError(_('Error running %s') % command, p.returncode)

    def start_build(self):
        self.server = ServerProxy(self.xmlrpc_report_url, allow_none = True)
        if self.verbose:
            self.server.verbose_timeout = True

        # close stdin
        sys.stdin.close()

        info = {}
        import socket
        un = os.uname()

        info['build_host'] = socket.gethostname()
        info['architecture'] = (un[0], un[2], un[4])

        distro = get_distro()
        if distro:
            info['distribution'] = distro

        info['module_set'] = self.config.moduleset

        try:
            self.build_id = self.server.start_build(info)
        except xmlrpclib.ProtocolError as e:
            if e.errcode == 403:
                print >> sys.stderr, _('ERROR: Wrong credentials, please check username/password')
                sys.exit(1)
            raise

        
        if self.verbose:
            s = _('Starting Build #%s') % self.build_id
            print s
            print '=' * len(s)
            print ''


    def end_build(self, failures):
        self.server.end_build(self.build_id, failures)
        if self.verbose:
            TerminalBuildScript.end_build(self, failures)


    def start_module(self, module):
        if self.verbose:
            print '\n%s' % t_bold + _('**** Starting module %s ****' % module) + t_reset
        self.server.start_module(self.build_id, module)
        self.current_module = module
        self.modulefp = StringIO()
        

    def end_module(self, module, failed):
        log = fix_encoding(self.modulefp.getvalue())
        self.modulefp = None
        self.server.end_module(self.build_id, module, compress_data(log), failed)

    def start_phase(self, module, phase):
        self.server.start_phase(self.build_id, module, phase)
        if self.verbose:
            TerminalBuildScript.start_phase(self, module, phase)
        self.phasefp = StringIO()


    def end_phase(self, module, phase, error):
        log = fix_encoding(self.phasefp.getvalue())
        self.phasefp = None

        if phase == 'test':
            if self.modules == {}:
                self.modules = jhbuild.moduleset.load_tests(self.config)

            if module in self.modules.modules.keys() \
                   and self.modules.modules[module].test_type == 'ldtp':
                self._upload_logfile(module)

        if isinstance(error, Exception):
            error = unicode(error)
        self.server.end_phase(self.build_id, module, phase, compress_data(log), error)

    def handle_error(self, module, phase, nextphase, error, altphases):
        '''handle error during build'''
        print 'FIXME: handle error! (failed build: %s: %s)' % (module, error)
        return 'fail'

    def _upload_ldtp_logfile (self, module):
        test_module = self.modules.modules[module]
        src_dir =  test_module.get_srcdir()
        if not os.path.exists (os.path.join(src_dir,'run.xml')):
            return
        logfile = test_module.get_ldtp_log_file (os.path.join(src_dir,'run.xml'))
        if not os.path.exists (logfile):
            return
        self._upload_logfile (module, logfile, 'application/x-ldtp+xml')

    def _upload_logfile (self, module, logfile, mimetype):
        log = open (logfile, 'r')
        basename = os.path.basename (logfile)
        self.server.attach_file (self.build_id, module, 'test', basename,
                                 compress_data(log.read()), mimetype)
        log.close()

BUILD_SCRIPT = AutobuildBuildScript



# jhbuild - a tool to ease building collections of source packages
# Copyright (C) 2008  Andy Wingo <wingo@pobox.com>
#
#   sxml.py: xml as s-expressions
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



"""
An s-expression syntax for XML documents, together with a serializer to
UTF-8.

Use like this:

>>> x = [sxml.h1, "text"]
>>> sxml_to_string (x)
"<h1>text</h1>"

>>> x = [sxml.a(href="about:blank", title="foo"), [sxml.i, "italics & stuff"]]
>>> sxml_to_string (x)
"<a href="about:blank" title="foo"><i>italics &amp; stuff</i></a>"
"""


__all__ = ['sxml', 'sxml_to_string']


# from Django, originally. used to make sure xml is utf-8.
def smart_str(s, encoding='utf-8', errors='strict'):
    # Returns a bytestring version of 's', encoded as specified in 'encoding'.
    if not isinstance(s, basestring):
        try:
            return str(s)
        except UnicodeEncodeError:
            return unicode(s).encode(encoding, errors)
    elif isinstance(s, unicode):
        return s.encode(encoding, errors)
    elif s and encoding != 'utf-8':
        return s.decode('utf-8', errors).encode(encoding, errors)
    else:
        return s

def quote(s):
    quoted = {'"': '&quot;',
              '&': '&amp;',
              '<': '&lt;',
              '>': '&gt;'}
    return ''.join([quoted.get(c,c) for c in s])

def sxml_to_string(expr):
    if not isinstance(expr, list):
        return smart_str(quote(expr))
    operator = expr[0]
    args = [sxml_to_string(arg) for arg in expr[1:]]
    return smart_str(operator(args))

class sxml:
    def __getattr__(self, attr):
        def _trans(k):
            table = {'klass': 'class'}
            return table.get(k, k)
        def tag(*targs, **kw):
            def render(args):
                return ('<%s%s>%s</%s>'
                        % (attr,
                           ''.join([' %s="%s"' % (_trans(k), quote(v))
                                    for k, v in kw.items()]),
                           '\n'.join(args),
                           attr))
            render.__name__ = attr
            if targs:
                return render(targs[0])
            else:
                return render
        # this only works with python2.4
        tag.__name__ = attr 
        return tag
sxml = sxml()



# jhbuild - a tool to ease building collections of source packages
# Copyright (C) 2001-2006  James Henstridge
#
#   __init__.py: infrastructure for plugging in version control systems
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

__all__ = [
    'Repository',
    'Branch',
    'register_repo_type',
    'get_repo_type',
    ]

__metaclass__ = type

from jhbuild.errors import FatalError, BuildStateError
import os

class Repository:
    """An abstract class representing a collection of modules."""

    # Attributes expected on the <repository> element for this type.
    # String values are passed as keyword arguments to the constructor.
    init_xml_attrs = []

    # URI of the moduleset where this repository is defined
    moduleset_uri = None

    def __init__(self, config, name):
        self.config = config
        self.name = name

    # Attributes expected on the <branch> element for this type.
    # String values are passed as keyword arguments to the branch() method.
    branch_xml_attrs = []

    def branch(self, name, **kwargs):
        """Returns a Branch object based on the given arguments."""
        raise NotImplementedError
    
    def branch_from_xml(self, name, branchnode, repositories, default_repo):
        kws = {}
        for attr in self.branch_xml_attrs:
            if branchnode.hasAttribute(attr):
                kws[attr.replace('-', '_')] = branchnode.getAttribute(attr)
        if branchnode.hasAttribute('id'):
            kws['branch_id'] = branchnode.getAttribute('id')
        return self.branch(name, **kws)

    def to_sxml(self):
        """Return an sxml representation of this repository."""
        raise NotImplementedError

    def get_sysdeps(self):
        return []


class Branch:
    """An abstract class representing a branch in a repository."""

    def __init__(self, repository, module, checkoutdir):
        self.repository = repository
        self.config = repository.config
        self.module = module
        self.checkoutdir = checkoutdir
        self.checkoutroot = self.config.checkoutroot

    def get_checkout_mode(self):
        checkout_mode = self.config.checkout_mode
        return self.config.module_checkout_mode.get(self.module, checkout_mode)
    checkout_mode = property(get_checkout_mode)

    def srcdir(self):
        """Return the directory where this branch is checked out."""
        raise NotImplementedError
    srcdir = property(srcdir)

    def branchname(self):
        """Return an identifier for this branch or None."""
        raise NotImplementedError
    branchname = property(branchname)

    def exists(self):
        """Return True if branch exists or False if not.

           May raise NotImplementedError if cannot check a given branch."""
        raise NotImplementedError

    def get_module_basename(self):
        # prevent basename() from returning empty strings on trailing '/'
        module = self.module.rstrip(os.sep)
        module = os.path.basename(module)
        # prune common filename extensions
        index = module.lower().rfind('.tar')
        if index > 0:
            return module[:index]
        if len(module) > 4 and module.lower().endswith('.zip'):
            return module[:-4]
        return module

    def get_checkoutdir(self):
        if self.checkout_mode == 'copy' and self.config.copy_dir:
            return os.path.join(self.config.copy_dir, self.get_module_basename())
        if self.checkoutdir:
            return os.path.join(self.checkoutroot, self.checkoutdir)
        if self.get_module_basename():
            return os.path.join(self.checkoutroot, self.get_module_basename())
        raise Exception('unable to get a valid checkout directory')

    def may_checkout(self, buildscript):
        if buildscript.config.nonetwork:
            return False
        return True

    def checkout(self, buildscript):
        """Checkout or update the given source directory.

        May raise CommandError or BuildStateError if a problem occurrs.
        """
        if self.checkout_mode in ('clobber', 'export'):
            self._wipedir(buildscript, self.srcdir)
            if self.checkout_mode == 'export':
                try:
                    self._export(buildscript)
                except NotImplementedError:
                    pass
                else:
                    return
            self._checkout(buildscript)
        elif self.checkout_mode in ('update', 'copy'):
            if self.checkout_mode == 'copy' and self.config.copy_dir:
                copydir = self.config.copy_dir
                if os.path.exists(self.get_checkoutdir()):
                    self._update(buildscript, self.config.copy_dir)
                else:
                    self._wipedir(buildscript, self.srcdir)
                    self._checkout(buildscript, copydir)
                self._copy(buildscript, copydir)
            else:
                if os.path.exists(self.get_checkoutdir()):
                    self._update(buildscript)
                else:
                    self._checkout(buildscript)

    def force_checkout(self, buildscript):
        """A more agressive version of checkout()."""
        self._wipedir(buildscript, self.srcdir)
        self.checkout(buildscript)

    def tree_id(self):
        """A string identifier for the state of the working tree."""
        raise NotImplementedError

    def _wipedir(self, buildscript, dir):
        if dir and dir != os.sep and os.path.exists(dir):
            buildscript.execute(['rm', '-rf', dir])

    def _export(self, buildscript):
        raise NotImplementedError

    def _copy(self, buildscript, copydir):
         module = self.module
         if self.checkoutdir:
             module = self.checkoutdir
         fromdir = os.path.join(copydir, os.path.basename(module))
         todir = os.path.join(self.config.checkoutroot, os.path.basename(module))
         if os.path.exists(todir):
             self._wipedir(buildscript, self.srcdir)
         buildscript.execute(['cp', '-R', fromdir, todir])

    def to_sxml(self):
        """Return an sxml representation of this checkout."""
        raise NotImplementedError

_repo_types = {}
def register_repo_type(name, repo_class):
    _repo_types[name] = repo_class

def get_repo_type(name):
    if name not in _repo_types:
        try:
            __import__('jhbuild.versioncontrol.%s' % name)
        except ImportError:
            pass
    if name not in _repo_types:
        raise FatalError(_('unknown repository type %s') % name)
    return _repo_types[name]



# jhbuild - a tool to ease building collections of source packages
# Copyright (C) 2001-2006  James Henstridge
#
#   tarball.py: rules for building tarballs
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

__metaclass__ = type

import sys
import logging

try:
    import hashlib
except ImportError:
    hashlib = None

from jhbuild.modtypes import register_module_type, get_dependencies, find_first_child_node_content

def parse_tarball(node, config, uri, repositories, default_repo):
    name = node.getAttribute('id')
    version = node.getAttribute('version')
    source_url = None
    source_size = None
    source_hash = None
    patches = []
    checkoutdir = None
    autogenargs = ''
    makeargs = ''
    makeinstallargs = ''
    supports_non_srcdir_builds = True
    makefile = 'Makefile'
    if node.hasAttribute('checkoutdir'):
        checkoutdir = node.getAttribute('checkoutdir')
    if node.hasAttribute('autogenargs'):
        autogenargs = node.getAttribute('autogenargs')
    if node.hasAttribute('makeargs'):
        makeargs = node.getAttribute('makeargs')
    if node.hasAttribute('makeinstallargs'):
        makeinstallargs = node.getAttribute('makeinstallargs')
    if node.hasAttribute('supports-non-srcdir-builds'):
        supports_non_srcdir_builds = \
            (node.getAttribute('supports-non-srcdir-builds') != 'no')
    if node.hasAttribute('makefile'):
        makefile = node.getAttribute('makefile')

    for childnode in node.childNodes:
        if childnode.nodeType != childnode.ELEMENT_NODE: continue
        if childnode.nodeName == 'source':
            source_url = childnode.getAttribute('href')
            if childnode.hasAttribute('size'):
                try:
                    source_size = int(childnode.getAttribute('size'))
                except ValueError:
                    logging.warning(
                            _('module \'%(module)s\' has invalid size attribute (\'%(size)s\')') % {
                                'module': name, 'size': childnode.getAttribute('size')})
            if childnode.hasAttribute('md5sum'):
                source_hash = 'md5:' + childnode.getAttribute('md5sum')
            if childnode.hasAttribute('hash') and hashlib:
                source_hash = childnode.getAttribute('hash')
        elif childnode.nodeName == 'patches':
            for patch in childnode.childNodes:
                if patch.nodeType != patch.ELEMENT_NODE: continue
                if patch.nodeName != 'patch': continue
                patchfile = patch.getAttribute('file')
                if patch.hasAttribute('strip'):
                    patchstrip = int(patch.getAttribute('strip'))
                else:
                    patchstrip = 0
                patches.append((patchfile, patchstrip))

    # for tarballs, don't ever pass --enable-maintainer-mode
    autogenargs = autogenargs.replace('--enable-maintainer-mode', '')

    dependencies, after, suggests, systemdependencies = get_dependencies(node)

    from autotools import AutogenModule
    from jhbuild.versioncontrol.tarball import TarballBranch, TarballRepository

    # create a fake TarballRepository, and give it the moduleset uri
    repo = TarballRepository(config, None, None)
    repo.moduleset_uri = uri

    branch = TarballBranch(repo, source_url, version, checkoutdir,
            source_size, source_hash, None)
    branch.patches = patches

    instance = AutogenModule(name, branch,
                             autogenargs, makeargs, makeinstallargs,
                             supports_non_srcdir_builds = supports_non_srcdir_builds,
                             skip_autogen = False, autogen_sh = 'configure',
                             makefile = makefile)
    instance.dependencies = dependencies
    instance.after = after
    instance.suggests = suggests
    instance.systemdependencies = systemdependencies
    instance.pkg_config = find_first_child_node_content(node, 'pkg-config')
    instance.config = config

    return instance

register_module_type('tarball', parse_tarball)



# jhbuild - a tool to ease building collections of source packages
# Copyright (C) 2001-2006  James Henstridge
# Copyright (C) 2003-2004  Seth Nickell
#
#   gui.py: the GTK interface for jhbuild
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import pygtk
pygtk.require('2.0')
import gtk

from jhbuild.commands import Command, register_command
import jhbuild.frontends


class cmd_gui(Command):
    doc = N_('Build targets from a GUI app')

    name = 'gui'
    usage_args = ''

    def run(self, config, options, args, help=None):
        # request GTK build script.
        config.buildscript = 'gtkui'

        build = jhbuild.frontends.get_buildscript(config)
        build.show()
        gtk.main()

register_command(cmd_gui)



# jhbuild - a tool to ease building collections of source packages
# Copyright (C) 2008  Frederic Peters
#
#   build.py: custom build pages
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import urllib, time

from twisted.web import html
from twisted.web.util import Redirect

from buildbot.status.web.build import BuildsResource, StatusResourceBuild
from buildbot.status.web.base import HtmlResource, make_row, make_stop_form, \
     css_classes, make_name_user_passwd_form

from step import JhStepsResource


class JhStatusResourceBuild(StatusResourceBuild):
    def getTitle(self, request):
        buildbot_service = request.site.buildbot_service
        builder_name = self.build_status.getBuilder().getName()
        for slave in buildbot_service.slaves:
            if builder_name.endswith(slave):
                slave_name = slave
                module_name = builder_name[:-len(slave)-1]
                break
        else:
            slave_name = None
            module_name = None
        status = self.getStatus(request)
        p = status.getProjectName()
        if slave_name:
            return '%s - %s - build #%s @ %s' % (
                    p, module_name, self.build_status.getNumber(), slave_name)
        return '%s - %s' % (p, builder_name)

    def body(self, req):
        b = self.build_status
        status = self.getStatus(req)
        projectURL = status.getProjectURL()
        projectName = status.getProjectName()
        data = ''

        if not b.isFinished():
            data += '<h2>Build In Progress</h2>'
            when = b.getETA()
            if when is not None:
                when_time = time.strftime('%H:%M:%S',
                                          time.localtime(time.time() + when))
                data += '<div>ETA %ds (%s)</div>\n' % (when, when_time)

            if self.build_control is not None:
                stopURL = urllib.quote(req.childLink("stop"))
                data += make_stop_form(stopURL, self.isUsingUserPasswd(req))

        if b.isFinished():
            results = b.getResults()
            data += "<h2>Results: "
            text = " ".join(b.getText())
            data += '<span class="%s">%s</span>' % (css_classes[results], text)
            data += '</h2>\n'
            if b.getTestResults():
                url = req.childLink("tests")
                data += "<h3><a href=\"%s\">test results</a></h3>\n" % url

        data += '<ul>\n'
        ss = b.getSourceStamp()
        data += " <li>SourceStamp: "
        data += " <ul>\n"
        if ss.branch:
            data += "  <li>Branch: %s</li>\n" % html.escape(ss.branch)
        if ss.revision:
            data += "  <li>Revision: %s</li>\n" % html.escape(str(ss.revision))
        if ss.patch:
            data += "  <li>Patch: YES</li>\n" # TODO: provide link to .diff
        if ss.changes:
            data += "  <li>Changes: see below</li>\n"
        if (ss.branch is None and ss.revision is None and ss.patch is None
            and not ss.changes):
            data += "  <li>build of most recent revision</li>\n"
        got_revision = None
        try:
            got_revision = b.getProperty("got_revision")
        except KeyError:
            pass
        if got_revision:
            got_revision = str(got_revision)
            if len(got_revision) > 40:
                got_revision = "[revision string too long]"
            data += "  <li>Got Revision: %s</li>\n" % got_revision
        data += " </ul>\n"

        data += "<li>Buildslave: %s</li>\n" % html.escape(b.getSlavename())
        data += "<li>Reason: %s</li>\n" % html.escape(b.getReason())

        if b.getLogs():
            data += "<li>Steps and Logfiles:\n"
            data += "<ol>\n"
            for s in b.getSteps():
                name = s.getName()
                data += (" <li><a href=\"%s\">%s</a> [%s]\n"
                         % (req.childLink("steps/%s" % urllib.quote(name)),
                            name,
                            " ".join(s.getText())))
                if s.getLogs():
                    data += "  <ul>\n"
                    for logfile in s.getLogs():
                        logname = logfile.getName()
                        logurl = req.childLink("steps/%s/logs/%s" %
                                               (urllib.quote(name),
                                                urllib.quote(logname)))
                        data += ("   <li><a href=\"%s\">%s</a></li>\n" %
                                 (logurl, logfile.getName()))
                    data += "  </ul>\n"
                data += " </li>\n"
            data += "</ol>\n"
            data += "</li>\n"

        data += "<li>Blamelist: "
        if list(b.getResponsibleUsers()):
            data += " <ol>\n"
            for who in b.getResponsibleUsers():
                data += "  <li>%s</li>\n" % html.escape(who)
            data += " </ol>\n"
        else:
            data += "no responsible users\n"
        data += '</li>'

        if ss.changes:
            data += "<li>All changes\n"
            data += "<ol>\n"
            for c in ss.changes:
                data += '<li class="changeset">' + c.asHTML() + '</li>\n'
            data += "</ol></li>\n"

        data += '</ul>'

        if b.isFinished() and self.builder_control is not None:
            data += "<h3>Resubmit Build:</h3>\n"
            # can we rebuild it exactly?
            exactly = (ss.revision is not None) or b.getChanges()
            if exactly:
                data += ("<p>This tree was built from a specific set of \n"
                         "source files, and can be rebuilt exactly</p>\n")
            else:
                data += ("<p>This tree was built from the most recent "
                         "revision")
                if ss.branch:
                    data += " (along some branch)"
                data += (" and thus it might not be possible to rebuild it \n"
                         "exactly. Any changes that have been committed \n"
                         "after this build was started <b>will</b> be \n"
                         "included in a rebuild.</p>\n")
            rebuildURL = urllib.quote(req.childLink("rebuild"))
            data += ('<form method="post" action="%s" class="command rebuild">\n'
                     % rebuildURL)
            data += make_name_user_passwd_form(self.isUsingUserPasswd(req))
            data += make_row("Reason for re-running build:",
                             "<input type='text' name='comments' />")
            data += '<input type="submit" value="Rebuild" />\n'
            data += '</form>\n'

        data += '</div>\n'

        return data

    def getChild(self, path, req):
        if path == 'steps':
            return JhStepsResource(self.build_status)
        return StatusResourceBuild.getChild(self, path, req)


class JhBuildsResource(BuildsResource):
    def getChild(self, path, req):
        if path == '':
            return Redirect('..')
        try:
            num = int(path)
        except ValueError:
            num = None
        if num is not None:
            build_status = self.builder_status.getBuild(num)
            if build_status:
                if self.builder_control:
                    build_control = self.builder_control.getBuild(num)
                else:
                    build_control = None
                return JhStatusResourceBuild(build_status, build_control,
                                           self.builder_control)

        return HtmlResource.getChild(self, path, req)




# jhbuild - a tool to ease building collections of source packages
# Copyright (C) 2001-2006  James Henstridge
#
#   svn.py: some code to handle various Subversion operations
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

__all__ = []
__metaclass__ = type

import os
import urlparse
import subprocess

from jhbuild.errors import CommandError, BuildStateError
from jhbuild.utils.cmds import get_output, check_version
from jhbuild.versioncontrol import Repository, Branch, register_repo_type
from jhbuild.commands.sanitycheck import inpath
from jhbuild.utils.sxml import sxml

import bzr, git

svn_one_five = None # is this svn 1.5

def _make_uri(repo, path):
    if repo[-1] != '/':
        return '%s/%s' % (repo, path)
    else:
        return repo + path

# Make sure that the urlparse module considers svn:// and svn+ssh://
# schemes to be netloc aware and set to allow relative URIs.
if 'svn' not in urlparse.uses_netloc:
    urlparse.uses_netloc.append('svn')
if 'svn' not in urlparse.uses_relative:
    urlparse.uses_relative.append('svn')
if 'svn+ssh' not in urlparse.uses_netloc:
    urlparse.uses_netloc.append('svn+ssh')
if 'svn+ssh' not in urlparse.uses_relative:
    urlparse.uses_relative.append('svn+ssh')

def get_svn_extra_env():
    # we run Subversion in the C locale, because Subversion localises
    # the key names in the output.  See bug #334678 for more info.
    #
    # Also we run it without the JHBuild LD_LIBRARY_PATH, as it can lead to
    # errors if it picks up jhbuilded gnutls library.  See bug #561191.
    return { 'LANGUAGE': 'C', 'LC_ALL': 'C', 'LANG': 'C',
             'LD_LIBRARY_PATH': os.environ.get('UNMANGLED_LD_LIBRARY_PATH') }

def get_info(filename):
    output = get_output(
        ['svn', 'info', filename], extra_env=get_svn_extra_env())
    ret = {}
    for line in output.splitlines():
        if ':' not in line: continue
        key, value = line.split(':', 1)
        ret[key.lower().strip()] = value.strip()
    return ret

def get_subdirs(url):
    print _("Getting SVN subdirs: this operation might be long...")
    output = get_output(
        ['svn', 'ls', '-R', url], extra_env=get_svn_extra_env())
    ret = []
    for line in output.splitlines():
        if not line[-1] == '/': continue
        ret.append (line)
    return ret

def get_externals(url):
    output = get_output(['svn', 'propget', 'svn:externals', url],
            extra_env=get_svn_extra_env())
    ret = {}
    for line in output.splitlines():
        if ' ' not in line: continue
        key, value = line.split(' ')
        ret[key.strip()] = value
    return ret

def get_uri(filename):
    try:
        info = get_info(filename)
    except CommandError:
        raise BuildStateError(_('could not get Subversion URI for %s')
                              % filename)
    if 'url' not in info:
        raise BuildStateError(_('could not parse "svn info" output for %s')
                              % filename)
    return info['url']

def call_with_info(proc, filename, *keys):
    info = get_info(filename)
    try:
        return proc(*[info[k] for k in keys])
    except KeyError:
        return None

class SubversionRepository(Repository):
    """A class used to work with a Subversion repository"""

    init_xml_attrs = ['href', 'trunk-template', 'branches-template', 'tags-template']

    def __init__(self, config, name, href, trunk_template=None, branches_template=None, tags_template=None):
        Repository.__init__(self, config, name)
        # allow user to adjust location of branch.
        self.href = config.repos.get(name, href)
        self.trunk_template = trunk_template or "%(module)s/trunk"
        self.branches_template = branches_template or "%(module)s/branches/%(branch)s"
        self.tags_template = tags_template or "%(module)s/tags/%(tag)s"
        self.svn_program = config.svn_program

    branch_xml_attrs = ['module', 'checkoutdir', 'revision', 'tag']

    def branch(self, name, module=None, checkoutdir=None, revision=None, tag=None):
        module_href = None
        if name in self.config.branches:
            if self.config.branches[name]:
                module_href = self.config.branches[name]
            else:
                module = None
                revision = None

        if not module:
            module = name

        if revision and not revision.isdigit():
            template = self.branches_template
        elif tag:
            template = self.tags_template
        else:
            template = self.trunk_template

        # Workarounds for people with hacked modulesets
        if "/" in module or "trunk" == module:
            template = "%(module)s"

        if module_href is None:
            template = self.href + template
            module_href = template % {
                'module': module,
                'branch': revision,
                'tag': tag,
            }

        if checkoutdir is None:
            checkoutdir = name

        # workaround for svn client not handling '..' in URL (#560246, #678869)
        if os.name != 'nt':
            splitted_href = list(urlparse.urlsplit(module_href))
            splitted_href[2] = os.path.abspath(splitted_href[2])
            module_href = urlparse.urlunsplit(splitted_href)

        if self.svn_program == 'bzr' and not revision:
            return bzr.BzrBranch(self, module_href, checkoutdir)
        elif self.svn_program == 'git-svn':
            return git.GitSvnBranch(self, module_href, checkoutdir, revision)
        else:
            return SubversionBranch(self, module_href, name, checkoutdir, revision)

    def to_sxml(self):
        return [sxml.repository(type='svn', name=self.name, href=self.href)]

    def get_sysdeps(self):
        return ['svn']


class SubversionBranch(Branch):
    """A class representing a Subversion branch"""

    def __init__(self, repository, module, module_name, checkoutdir, revision):
        Branch.__init__(self, repository, module, checkoutdir)
        self.module_name = module_name
        self.revision = revision

    def srcdir(self):
        if self.checkoutdir:
            return os.path.join(self.checkoutroot, self.checkoutdir)
        else:
            return os.path.join(self.checkoutroot,
                                os.path.basename(self.module))
    srcdir = property(srcdir)

    def branchname(self):
        return self.revision
    branchname = property(branchname)

    def exists(self):
        try:
            get_output(['svn', 'ls', self.module], extra_env={
                'LD_LIBRARY_PATH': os.environ.get('UNMANGLED_LD_LIBRARY_PATH'),
                })
            return True
        except:
            return False

    def _export(self, buildscript):
        cmd = ['svn', 'export', self.module]

        if self.checkoutdir:
            cmd.append(self.checkoutdir)

        if self.revision and self.revision.isdigit():
            cmd.extend(['-r', '%s' % self.revision])
        elif self.config.sticky_date:
            cmd.extend(['-r', '{%s}' % self.config.sticky_date])

        buildscript.execute(cmd, 'svn', cwd=self.checkoutroot,
                extra_env = get_svn_extra_env())
    
    def _checkout(self, buildscript, copydir=None):
        cmd = ['svn', 'checkout', self.module]

        if self.checkoutdir:
            cmd.append(self.checkoutdir)

        if not self.config.interact:
            cmd.append('--non-interactive')

        if self.revision and self.revision.isdigit():
            cmd.extend(['-r', '%s' % self.revision])
        elif self.config.sticky_date:
            cmd.extend(['-r', '{%s}' % self.config.sticky_date])

        if copydir:
            buildscript.execute(cmd, 'svn', cwd=copydir,
                    extra_env=get_svn_extra_env())
        else:
            buildscript.execute(cmd, 'svn', cwd=self.config.checkoutroot,
                    extra_env=get_svn_extra_env())

    def _update(self, buildscript, copydir=None):
        opt = []
        if copydir:
            outputdir = os.path.join(copydir, os.path.basename(self.srcdir))
        else:
            outputdir = self.srcdir

        opt = []

        if not self.config.interact:
            opt.append('--non-interactive')

        if self.revision and self.revision.isdigit():
            opt.extend(['-r', '%s' % self.revision])
        elif self.config.sticky_date:
            opt.extend(['-r', '{%s}' % self.config.sticky_date])

        # Subversion 1.5 has interactive behaviour on conflicts; check version
        # and add appropriate flags to get back 1.4 behaviour.
        global svn_one_five
        if svn_one_five is None:
            svn_one_five = check_version(['svn', '--version'], r'svn, version ([\d.]+)', '1.5',
                    extra_env=get_svn_extra_env())

        if svn_one_five is True:
            opt.extend(['--accept', 'postpone'])

        uri = get_uri(outputdir)

        if urlparse.urlparse(uri)[:2] != urlparse.urlparse(self.module)[:2]:
            # server and protocol changed, probably because user changed
            # svnroots[] config variable.
            new_uri = urlparse.urlunparse(
                    urlparse.urlparse(self.module)[:2] + urlparse.urlparse(uri)[2:])
            cmd = ['svn', 'switch', '--relocate', uri, new_uri, '.']
            buildscript.execute(cmd, 'svn', cwd=outputdir,
                    extra_env=get_svn_extra_env())

        # if the URI doesn't match, use "svn switch" instead of "svn update"
        if get_uri(outputdir) != self.module:
            cmd = ['svn', 'switch'] + opt + [self.module]
        else:
            cmd = ['svn', 'update'] + opt + ['.']

        buildscript.execute(cmd, 'svn', cwd=outputdir,
                extra_env=get_svn_extra_env())

        try:
            self._check_for_conflicts()
        except CommandError:
            # execute svn status so conflicts are displayed
            buildscript.execute(['svn', 'status'], 'svn', cwd=self.srcdir,
                    extra_env=get_svn_extra_env())
            raise

    def _check_for_conflicts(self):
        kws = {}
        kws['cwd'] = self.srcdir
        kws['env'] = os.environ.copy()
        extra_env = get_svn_extra_env()
        kws['env'].update(extra_env)
        try:
            output = subprocess.Popen(['svn', 'info', '-R'],
                    stdout = subprocess.PIPE, **kws).communicate()[0]
        except OSError as e:
            raise CommandError(str(e))
        if '\nConflict' in output:
            raise CommandError(_('Error checking for conflicts'))

    def checkout(self, buildscript):
        if not inpath('svn', os.environ['PATH'].split(os.pathsep)):
            raise CommandError(_('%s not found') % 'svn')
        Branch.checkout(self, buildscript)

    def tree_id(self):
        if not os.path.exists(self.srcdir):
            return None
        try:
            if self.config.checkout_mode == 'export':
                info = get_info(self.module)
            else:
                info = get_info(self.srcdir)
        except CommandError:
            return None
        try:
            url = info['url']
            root = info['repository root']
            uuid = info['repository uuid']
            rev = info['last changed rev']
        except KeyError:
            return None

        # get the path within the repository
        assert url.startswith(root)
        path = url[len(root):]
        while path.startswith('/'):
            path = path[1:]

        return '%s,%s,%s' % (uuid.lower(), rev, path)

    def to_sxml(self):
        return (call_with_info(lambda rev:
                                   [sxml.branch(repo=self.repository.name,
                                                module=self.module,
                                                revision=rev)],
                               self.srcdir, 'last changed rev')
                or [sxml.branch(repo=self.repository.name,
                                module=self.module)])


register_repo_type('svn', SubversionRepository)



#! /usr/bin/env python2
#
# hg-update - pull and update a mercurial repository
#
# Copyright (C) 2007  Marco Barisione <marco@barisione.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import os
import sys
import re

from subprocess import Popen, call, PIPE, STDOUT

def get_parent():
    hg = Popen(['hg', 'parents', '--template', '{rev}'], stdout=PIPE)
    try:
        return hg.stdout.read().split()[0]
    except IndexError:
        # handle parentless revisions
        return ''

def pull():
    ret = call(['hg', 'pull'])
    return ret == 0

def update():
    env = dict(os.environ)
    env['HGMERGE'] = '/bin/false'
    env['LANG'] = 'C'
    hg = Popen(['hg', 'update', '--noninteractive'], stdout=PIPE,
               stderr=STDOUT, env=env)
    out = hg.communicate()[0]
    if hg.returncode != 0:
        # Use CVS-like format for conflicts.
        out = re.sub('merging (.*) failed!', r'C \1', out)
        index = out.find('You can redo the full merge using:')
        # Remove the instructions to redo the full merge as we are
        # going to revert the update.
        if index != -1:
            out = out[:index]
    print out
    return hg.returncode == 0

def undo_update(parent):
    print 'Update failed, updating to parent revision'
    env = dict(os.environ)
    env['HGMERGE'] = 'false'
    hg = call(['hg', 'update', '--noninteractive', '-q', parent], env=env)

def pull_and_update():
    parent = get_parent()
    if not pull():
        return False
    if update():
        return True
    else:
        undo_update(parent)
        return False

if __name__ == '__main__':
    ret = False
    try:
        ret = pull_and_update()
    except OSError as e:
        print '%s: %s' % (sys.argv[0], e)

    if ret:
        exit_code = 0
    else:
        exit_code = 1
    sys.exit(exit_code)




# jhbuild - a tool to ease building collections of source packages
# Copyright (C) 2001-2006  James Henstridge
# Copyright (C) 2007  Marco Barisione <marco@barisione.org>
#
#   hg.py: some code to handle various Mercurial operations
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

__all__ = []
__metaclass__ = type

import os
import sys
import urlparse
from subprocess import Popen, PIPE

from jhbuild.errors import FatalError, CommandError
from jhbuild.versioncontrol import Repository, Branch, register_repo_type
from jhbuild.commands.sanitycheck import inpath

class HgRepository(Repository):
    """A class representing a Mercurial repository.

    Note that this is just the parent directory for a bunch of hg
    branches, making it easy to switch to a mirror URI.
    """

    init_xml_attrs = ['href']

    def __init__(self, config, name, href):
        Repository.__init__(self, config, name)
        # allow user to adjust location of branch.
        self.href = config.repos.get(name, href)

    branch_xml_attrs = ['module', 'checkoutdir']

    def branch(self, name, module=None, checkoutdir=None):
        if name in self.config.branches:
            module = self.config.branches[name]
            if not module:
                raise FatalError(_('branch for %(name)s has wrong override, check your %(filename)s') % \
                                   {'name'     : name,
                                    'filename' : self.config.filename})
        else:
            if module is None:
                module = name
            if not self.href.startswith('ssh://'):
                module = urlparse.urljoin(self.href, module)
            else:
                module = self.href + module
        return HgBranch(self, module, checkoutdir)

    def get_sysdeps(self):
        return ['hg']


class HgBranch(Branch):
    """A class representing a Mercurial branch."""

    def srcdir(self):
        if self.checkoutdir:
            return os.path.join(self.checkoutroot, self.checkoutdir)
        else:
            return os.path.join(self.checkoutroot,
                                os.path.basename(self.module))
    srcdir = property(srcdir)

    def branchname(self):
        return None
    branchname = property(branchname)

    def _checkout(self, buildscript):
        cmd = ['hg', 'clone', self.module]
        if self.checkoutdir:
            cmd.append(self.checkoutdir)

        if self.config.sticky_date:
            raise FatalError(_('date based checkout not yet supported\n'))

        buildscript.execute(cmd, 'hg', cwd=self.checkoutroot)

    def _update(self, buildscript):
        if self.config.sticky_date:
            raise FatalError(_('date based checkout not yet supported\n'))
        if os.path.exists(SRCDIR):
            hg_update_path = os.path.join(SRCDIR, 'scripts', 'hg-update.py')
        else:
            hg_update_path = os.path.join(PKGDATADIR, 'hg-update.py')
        hg_update_path = os.path.normpath(hg_update_path)
        buildscript.execute([hg_update_path], 'hg', cwd=self.srcdir)

    def checkout(self, buildscript):
        if not inpath('hg', os.environ['PATH'].split(os.pathsep)):
            raise CommandError(_('%s not found') % 'hg')
        Branch.checkout(self, buildscript)

    def get_revision_id(self):
        # Return the id of the tip, see bug #313997.
        try:
            hg = Popen(['hg', 'ti', '--template', '{node}'], stdout=PIPE,
                       cwd=self.srcdir)
        except OSError as e:
            raise CommandError(str(e))
        return hg.stdout.read().strip()

    def tree_id(self):
        if not os.path.exists(self.srcdir):
            return None
        return self.get_revision_id()

register_repo_type('hg', HgRepository)



# jhbuild - a tool to ease building collections of source packages
# Copyright (C) 2001-2006  James Henstridge
#
#   tinderbox.py: build logic for a non-interactive reporting build
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import os
import time
import subprocess
import locale
import logging
import codecs
import sys

from jhbuild.main import _encoding
from jhbuild.utils import cmds
from jhbuild.utils import sysid
from jhbuild.errors import CommandError, FatalError
import buildscript
import commands

index_header = '''<html>
  <head>
    <meta http-equiv="Content-Type" content="text/html;charset=%(charset)s">
    <title>JHBuild Results</title>
    <style type="text/css">
      .section {
        margin-after: 1.5em;
      }
      .success {
        color: black;
        background-color: #afa;
      }
      .failure {
        color: black;
        background-color: #faa;
      }
    </style>
  </head>
  <body>
    <h1>JHBuild Results</h1>

    <div class="section">
    <h2>Platform</h2>
    %(buildplatform)s
    </div>

    <div class="section">
    <h2>Summary</h2>
    <table>
      <tr>
        <th>Time</th>
        <th>Module</th>
        <th>Phases</th>
        <th>Status</th>
      </tr>
'''
index_footer = '''
    </table>
    </div>

    %(failures)s
  </body>
</html>
'''

buildlog_header = '''<html>
  <head>
    <meta http-equiv="Content-Type" content="text/html;charset=%(charset)s">
    <title>%(module)s Build Log</title>
    <style type="text/css">
      pre {
        /* unfortunately, white-space: pre-wrap is not widely supported ... */
        white-space: -moz-pre-wrap; /* Mozilla based browsers */
        white-space: -pre-wrap;     /* Opera 4 - 6 */
        white-space: -o-pre-wrap;   /* Opera >= 7 */
        white-space: pre-wrap;      /* CSS3 */
        word-wrap: break-word;      /* IE 5.5+ */
      }
      .message {
        font-size: larger;
      }
      .timestamp{
        font-size: smaller;
        font-style: italic;
      }
      .command {
        color: blue;
      }
      .conflict {
        color: red;
      }
      .critical {
        background-color: red;
      }
      .error {
        color: red;
      }
      .warning {
        color: darkgreen;
        font-size: smaller;
      }
      .info {
        color: darkgreen;
        font-size: smaller;
      }
      .debug {
        color: gray;
        font-size: smaller;
      }
      .note {
        background-color: #FFFF66;
      }
    </style>
  </head>
  <body>
    <h1>%(module)s Build Log</h1>
'''
buildlog_footer = '''
  </body>
</html>
'''

def escape(string):
    if type(string) is not unicode:
        string = unicode(string, _encoding, 'replace')
    string = string.replace('&', '&amp;').replace('<','&lt;').replace(
            '>','&gt;').replace('\n','<br/>').replace(
            '\t','&nbsp;&nbsp;&nbsp;&nbsp;')
    return string

class LoggingFormatter(logging.Formatter):
    def __init__(self):
        logging.Formatter.__init__(self, '<div class="%(levelname)s">'
                                   '%(message)s</div>')

class TinderboxBuildScript(buildscript.BuildScript):
    triedcheckout = None

    def __init__(self, config, module_list, module_set=None):
        buildscript.BuildScript.__init__(self, config, module_list, module_set=module_set)
        self.indexfp = None
        self.modulefp = None

        for handle in logging.getLogger().handlers:
            handle.setFormatter(LoggingFormatter())

        self.outputdir = os.path.abspath(config.tinderbox_outputdir)
        if not os.path.exists(self.outputdir):
            os.makedirs(self.outputdir)

        os.environ['TERM'] = 'dumb'

        self.charset = _encoding

    def timestamp(self):
        tm = time.time()
        s = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(tm))
        msecs = max(int((tm - int(tm)) * 1000), 0)
        return '%s.%03d' % (s, msecs)

    def message(self, msg, module_num=-1):
        '''Display a message to the user'''
        if self.modulefp:
            self.modulefp.write('<div><b class="message">%s</b> '
                                '<span class="timestamp">%s</span></div>\n'
                                % (escape(msg), self.timestamp()))
        else:
            # do something with messages outside of builds of module builds
            pass

    def set_action(self, action, module, module_num=-1, action_target=None):
        if module_num == -1:
            module_num = self.module_num
        if not action_target:
            action_target = module.name
        self.message('%s %s' % (action, action_target), module_num)

    def execute(self, command, hint=None, cwd=None, extra_env=None):
        assert self.modulefp, 'not currently building a module'

        kws = {
            'close_fds': True
            }
        print_args = {}
        if cwd:
            print_args['cwd'] = cwd
        else:
            print_args['cwd'] = os.getcwd()

        self.modulefp.write('<pre>')
        if isinstance(command, (str, unicode)):
            kws['shell'] = True
            print_args['command'] = command
        else:
            print_args['command'] = ' '.join(command)

        if self.config.print_command_pattern:
            try:
                commandstr = self.config.print_command_pattern % print_args
                self.modulefp.write('<span class="command">%s</span>\n'
                                    % escape(commandstr))
            except TypeError as e:
                raise FatalError('\'print_command_pattern\' %s' % e)
            except KeyError as e:
                raise FatalError(_('%(configuration_variable)s invalid key'
                                   ' %(key)s' % \
                                   {'configuration_variable' :
                                        '\'print_command_pattern\'',
                                    'key' : e}))

        kws['stdin'] = subprocess.PIPE
        kws['stdout'] = subprocess.PIPE
        kws['stderr'] = subprocess.PIPE
        if hint == 'cvs':
            def format_line(line, error_output, fp=self.modulefp):
                if line[-1] == '\n': line = line[:-1]
                if line.startswith('C '):
                    fp.write('<span class="conflict">%s</span>\n'
                                        % escape(line))
                else:
                    fp.write('%s\n' % escape(line))
            kws['stderr'] = subprocess.STDOUT
        else:
            def format_line(line, error_output, fp=self.modulefp):
                if line[-1] == '\n': line = line[:-1]
                if error_output:
                    fp.write('<span class="error">%s</span>\n'
                                        % escape(line))
                else:
                    fp.write('%s\n' % escape(line))

        if cwd is not None:
            kws['cwd'] = cwd

        if extra_env is not None:
            kws['env'] = os.environ.copy()
            kws['env'].update(extra_env)

        command = self._prepare_execute(command)

        try:
            p = subprocess.Popen(command, **kws)
        except OSError as e:
            self.modulefp.write('<span class="error">Error: %s</span>\n'
                                % escape(str(e)))
            raise CommandError(str(e))
        cmds.pprint_output(p, format_line)
        self.modulefp.write('</pre>\n')
        self.modulefp.flush()
        if p.returncode != 0:
            raise CommandError('Error running %s' % print_args['command'],
                               p.returncode)

    def start_build(self):
        assert self.outputdir

        # close stdin
        sys.stdin.close()

        info = []
        import socket
        un = os.uname()

        info.append(('Build Host', socket.gethostname()))
        info.append(('Architecture', '%s %s (%s)' % (un[0], un[2], un[4])))
        info.append(('System', sysid.get_pretty_name()))
        info.append(('Module Set', self.config.moduleset))
        info.append(('Start Time', self.timestamp()))

        buildplatform = '<table>\n'
        for (key, val) in info:
            buildplatform += '<tr><th align="left">%s</th><td>%s</td></tr>\n' \
                             % (key, val)
        buildplatform += '</table>\n'
        
        self.indexfp = codecs.open(os.path.join(self.outputdir, 'index.html'),
                'w', encoding=self.charset, errors='xmlcharrefreplace')

        self.indexfp.write(index_header % { 'buildplatform': buildplatform,
                                            'charset': self.charset })
        self.indexfp.flush()

    def end_build(self, failures):
        self.indexfp.write('<tr>'
                           '<td>%s</td>'
                           '<td>finish</td>'
                           '</tr>\n' % self.timestamp())
        if failures:
            info = '<p>The following modules failed to build</p>\n'
            info += '<blockquote>\n'
            info += ', '.join(failures)
            info += '</blockquote>\n'
        else:
            info = ''
        self.indexfp.write(index_footer % { 'failures': info })
        self.indexfp.close()
        self.indexfp = None

    def start_module(self, module):
        self.modulefilename='%s.html' % module.replace('/','_')
        self.indexfp.write('<tr>'
                           '<td>%s</td>'
                           '<td><a href="%s">%s</a></td>'
                           '<td>\n' % (self.timestamp(), self.modulefilename,
                                       module))
        self.modulefp = codecs.open(
                os.path.join(self.outputdir, self.modulefilename), 'w',
                encoding=self.charset, errors='xmlcharrefreplace')

        for handle in logging.getLogger().handlers:
            if isinstance(handle, logging.StreamHandler):
                handle.stream = self.modulefp

        self.modulefp.write(buildlog_header % { 'module': module,
                                                'charset': self.charset })
    def end_module(self, module, failed):
        if failed:
            self.message('Failed')
        else:
            self.message('Succeeded')
        self.modulefp.write(buildlog_footer)
        self.modulefp.close()
        self.modulefp = None
        self.indexfp.write('</td>\n')
        if failed:
            help_html = ''
            if self.config.help_website and self.config.help_website[1]:
                try:
                    help_url = self.config.help_website[1] % {'module' : module}
                    help_html = ' <a href="%s">(help)</a>' % help_url
                except TypeError as e:
                    raise FatalError('"help_website" %s' % e)
                except KeyError as e:
                    raise FatalError(_('%(configuration_variable)s invalid key'
                                       ' %(key)s' % \
                                       {'configuration_variable' :
                                            '\'help_website\'',
                                        'key' : e}))

            self.indexfp.write('<td class="failure">failed%s</td>\n' %
                               help_html)
        else:
            self.indexfp.write('<td class="success">ok</td>\n')
        self.indexfp.write('</tr>\n\n')
        self.indexfp.flush()

    def start_phase(self, module, phase):
        self.modulefp.write('<a name="%s"></a>\n' % phase)
    def end_phase(self, module, phase, error):
        if error:
            self.indexfp.write('<a class="failure" title="%s" href="%s#%s">%s</a>\n'
                               % (error, self.modulefilename, phase, phase))
        else:
            self.indexfp.write('<a class="success" href="%s#%s">%s</a>\n'
                               % (self.modulefilename, phase, phase))
        self.indexfp.flush()

    def handle_error(self, module, phase, nextphase, error, altphases):
        '''handle error during build'''
        self.message('error during stage %s of %s: %s' % (phase, module.name,
                                                          error))
        if self.config.trycheckout:
            if self.triedcheckout is None and altphases.count('configure'):
                self.triedcheckout = 'configure'
                self.message(_('automatically retrying configure'))
                return 'configure'
            elif self.triedcheckout == 'configure' and \
            altphases.count('force_checkout'):
                self.triedcheckout = 'done'
                self.message(_('automatically forcing a fresh checkout'))
                return 'force_checkout'
        self.triedcheckout = None

        if (self.modulefp and self.config.help_website and
            self.config.help_website[0] and self.config.help_website[1]):
            try:
                help_url = self.config.help_website[1] % \
                               {'module' : module.name}
                self.modulefp.write('<div class="note">The %(name)s website may'
                                    ' have suggestions on how to resolve some'
                                    ' build errors. Visit'
                                    ' <a href="%(url)s">%(url)s</a>'
                                    ' for more information.</div>'
                                    % {'name' : self.config.help_website[0],
                                       'url'  : help_url})
            except TypeError as e:
                raise FatalError('"help_website" %s' % e)
            except KeyError as e:
                raise FatalError(_('%(configuration_variable)s invalid key'
                                   ' %(key)s' % \
                                   {'configuration_variable' :
                                        '\'help_website\'',
                                    'key' : e}))
        return 'fail'

BUILD_SCRIPT = TinderboxBuildScript



# jhbuild - a tool to ease building collections of source packages
# Copyright (C) 2008  Frederic Peters
#
#   changes.py: custom changes pages
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import time

from buildbot.status.web.base import HtmlResource
from twisted.web.util import Redirect
from twisted.web.error import NoResource

class ChangesResource(HtmlResource):
    def getChild(self, path, req):
        if path == '':
            return Redirect('..')
        num = int(path)
        c = self.getStatus(req).getChange(num)
        if not c:
            return NoResource("No change number '%d'" % num)
        return ChangeResource(c)


class ChangeResource(HtmlResource):
    def __init__(self, change):
        self.change = change

    def getTitle(self, request):
        status = self.getStatus(request)
        p = status.getProjectName()
        if len(self.change.revision) == 40:
            return '%s - %s - commit %s' % (p, self.change.project, self.change.revision[:8])
        else:
            return '%s - %s - revision #%s' % (p, self.change.project, int(self.change.revision))

    def body(self, request):
        data = '<div class="changeset">\n'
        data += '<ul>'
        if self.change.project:
            data += '<li>Project: <a href="../%s">%s</a></li>\n' % (
                    self.change.project, self.change.project)
        if self.change.who:
            data += '<li>Author: <strong class="author">%s</strong></li>\n' % self.change.who
        if self.change.when:
            data += '<li>Date: <strong class="date">%s</strong></li>\n' % time.strftime(
                    '%a %d %b %Y %H:%M:%S', time.localtime(self.change.when))
        if self.change.files:
            data += '<li>Files:<ul>\n'
            for f in self.change.files:
                data += '<li><tt>%s</tt></li>\n' % f
            data += '</ul></li>\n'
        data += '</ul>\n'
        if self.change.comments:
            data += '<pre>'
            data += self.change.comments
            data += '</pre>\n'

        if self.change.revision:
            if len(self.change.revision) == 40:
                # git commit
                if hasattr(self.change, 'git_module_name'):
                    git_module_name = self.change.git_module_name
                else:
                    git_module_name = self.change.project
                link = 'https://git.gnome.org/browse/%s/commit/?id=%s' % (
                        git_module_name, self.change.revision)
                data += '<p>View in GNOME cgit: <a href="%s">%s commit %s</a></dd>\n' % (
                        link, git_module_name, self.change.revision[:8])
            else:
                link = 'http://svn.gnome.org/viewvc/%s?view=revision&revision=%s' % (
                        self.change.project, self.change.revision)
                data += '<p>View in GNOME ViewVC: <a href="%s">%s r%s</a></dd>\n' % (
                        link, self.change.project, self.change.revision)

        data += '</div>'
        return data




# jhbuild - a tool to ease building collections of source packages
# Copyright (C) 2008  apinheiro@igalia.com, John Carr, Frederic Peters
#
#   __init__.py: custom buildbot web pages
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import os

from twisted.web import server, static, resource
from buildbot.status.web.base import HtmlResource, ITopBox, build_get_class
from buildbot import interfaces, util
from buildbot.status.builder import SUCCESS, WARNINGS, FAILURE, EXCEPTION
from buildbot.status.web.baseweb import WebStatus

from waterfall import JhWaterfallStatusResource
from changes import  ChangesResource
from builder import JhBuildersResource
from bot import JhBuildbotsResource


def content(self, request):
    """
    We want to give /all/ HTMLResource objects this replacement content method
    Monkey patch :)
    """
    s = request.site.buildbot_service
    data = s.template
    data = data.replace('@@GNOME_BUILDBOT_TITLE@@', self.getTitle(request))
    data = data.replace('@@GNOME_BUILDBOT_BODY@@', self.body(request))
    return data
HtmlResource.content = content


class ListOfModules(resource.Resource):
    def render(self, request):
        data = self.content(request)
        request.setHeader('content-type', 'text/plain')
        if request.method == 'HEAD':
            request.setHeader('content-length', len(data))
            return ''
        return str(data)

    def content(self, request):
        parent = request.site.buildbot_service
        return '\n'.join(parent.modules)


class ProjectsSummary(HtmlResource):

    MAX_PROJECT_NAME = 25

    def getTitle(self, request):
        status = self.getStatus(request)
        p = status.getProjectName()
        return p

    def body(self, request):
        parent = request.site.buildbot_service
        status = self.getStatus(request)

        result = ''
        result += '<table class="ProjectSummary">\n'

        # Headers
        slave_status = {}
        for slave in parent.slaves:
            for module in parent.modules:
                builder = status.getBuilder("%s-%s" % (module, slave))
                state, builds = builder.getState()
                if state == 'offline':
                    slave_status[slave] = ('offline', [])
                    break
                elif state == 'building':
                    if slave in slave_status:
                        modules = slave_status[slave][1] or []
                        slave_status[slave] = (state, modules + [module])
                    else:
                        slave_status[slave] = (state, [module])
            else:
                if not slave in slave_status:
                    slave_status[slave] = ('idle', None)

        if type(parent.moduleset) is list:
            moduleset = ', '.join(parent.moduleset)
        else:
            moduleset = parent.moduleset
        result += '<thead><tr><td>&nbsp;</td><th>' + moduleset + '</td>'
        for name in parent.slaves:
            if len(name) > 25:
                name = name[:25] + '(...)'
            klass, modules = slave_status.get(name)
            if klass == 'building':
                title = 'Building %s' % ', '.join(modules)
            else:
                title = klass
            result += '<th class="%s" title="%s"><a href="bots/%s">%s</a></th>' % (
                    klass, title, name, name)
        result += '</tr>'
        thead = result
        # stop it here as a row with totals will be added here once every rows
        # have been handled

        # Contents
        result = '<tbody>'

        slave_results = {}
        for slave in parent.slaves:
            slave_results[slave] = [0, 0, 0]

        for module in parent.modules:
            result += '<tr>'
            result += '<td class="feed"><a href="%s/atom">' % module
            result += '<img src="/feed.png" alt="Atom"></a></td>\n'
            result += '<th><a href="%s">%s</a></td>' % (module, module)

            for slave in parent.slaves:
                builder = status.getBuilder("%s-%s" % (module, slave))
                box = ITopBox(builder).getBox(request)
                lastbuild = ''
                for bt in box.text:
                    if bt == 'successful' or bt == 'failed':
                        lastbuild = bt

                if lastbuild == 'successful':
                    last_build = builder.getLastFinishedBuild()
                    if last_build:
                        class_ = build_get_class(last_build)
                    else:
                        class_ = 'success'
                    lastbuild_label = 'Success'
                    if last_build and class_:
                        # use a different class/label if make check failed
                        steps = last_build.getSteps()
                        for step in reversed(steps):
                            if step.name.split()[-1] == 'check':
                                if step.results == WARNINGS:
                                    # make check failed
                                    class_ = 'failedchecks'
                                    lastbuild_label = 'Failed Checks'
                                break
                elif lastbuild == 'failed':
                    lastbuild_label = 'Failed'
                    last_build = builder.getLastFinishedBuild()
                    if last_build:
                        class_ = build_get_class(last_build)
                    else:
                        class_ = 'failure'
                else:
                    class_ = ''
                    lastbuild_label = lastbuild
                state, builds = builder.getState()
                if state == 'building':
                    result += '<td class="%s">%s</td>' % (state, state)
                else:
                    result += '<td class="%s">%s</td>' % (class_, lastbuild_label)
                
                if lastbuild in ('failed', 'successful'):
                    slave_results[slave][2] += 1
                    if class_ == 'failedchecks':
                        slave_results[slave][1] += 1
                    elif lastbuild == 'successful':
                        slave_results[slave][0] += 1
                        slave_results[slave][1] += 1

            result += '</tr>\n'
        result += '</tbody>\n'
        result += '<tfoot><tr class="totals"><td colspan="2"></td>'
        thead += '<tr class="totals"><td colspan="2"></td>'
        for slave in parent.slaves:
            td = '<td><span title="Successful builds">%s</span> '\
                      '<span title="(ignoring test suites failures)">(%s)</span> / '\
                      '<span title="Total">%s</span></td>' % tuple(slave_results[slave])
            thead += td
            result += td
        thead += '</tr>\n</thead>\n'
        result += '</tr></tfoot>\n'
        result += '</table>'

        return thead+result

class JHBuildWebStatus(WebStatus):

    def __init__(self, moduleset, modules, slaves, *args, **kwargs):
        WebStatus.__init__(self, *args, **kwargs)
        self.moduleset = moduleset
        self.modules = modules
        self.slaves = slaves

        # set up the per-module waterfalls
        for module in self.modules:
            self.putChild(module, JhWaterfallStatusResource(categories=[module]))

        # set the summary homepage
        self.putChild("", ProjectsSummary())

        # set custom changes pages
        self.putChild('changes', ChangesResource())
        self.putChild('builders', JhBuildersResource())
        self.putChild('bots', JhBuildbotsResource())

        # and more pages
        self.putChild('modules.txt', ListOfModules())

    def setupSite(self):
        WebStatus.setupSite(self)

        # load the template into memory
        self.template = open(os.path.join(self.parent.basedir, "template.html")).read()



# jhbuild - a tool to ease building collections of source packages
# Copyright (C) 2001-2006  James Henstridge
# Copyright (C) 2007-2008  Marc-Andre Lureau
#
#   git.py: some code to handle various GIT operations
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

__all__ = []
__metaclass__ = type

import os
import stat
import urlparse
import subprocess
import re
import urllib
import sys
import logging

from jhbuild.errors import FatalError, CommandError
from jhbuild.utils.cmds import get_output, check_version
from jhbuild.versioncontrol import Repository, Branch, register_repo_type
import jhbuild.versioncontrol.svn
from jhbuild.commands.sanitycheck import inpath
from jhbuild.utils.sxml import sxml

# Make sure that the urlparse module considers git:// and git+ssh://
# schemes to be netloc aware and set to allow relative URIs.
if 'git' not in urlparse.uses_netloc:
    urlparse.uses_netloc.append('git')
if 'git' not in urlparse.uses_relative:
    urlparse.uses_relative.append('git')
if 'git+ssh' not in urlparse.uses_netloc:
    urlparse.uses_netloc.append('git+ssh')
if 'git+ssh' not in urlparse.uses_relative:
    urlparse.uses_relative.append('git+ssh')
if 'ssh' not in urlparse.uses_relative:
    urlparse.uses_relative.append('ssh')

def get_git_extra_env():
    # we run git without the JHBuild LD_LIBRARY_PATH and PATH, as it can
    # lead to errors if it picks up jhbuilded libraries, such as nss
    return { 'LD_LIBRARY_PATH': os.environ.get('UNMANGLED_LD_LIBRARY_PATH'),
             'PATH': os.environ.get('UNMANGLED_PATH')}

def get_git_mirror_directory(mirror_root, checkoutdir, module):
    """Calculate the mirror directory from the arguments and return it."""
    mirror_dir = os.path.join(mirror_root, checkoutdir or
            os.path.basename(module))
    if mirror_dir.endswith('.git'):
        return mirror_dir
    else:
        return mirror_dir + '.git'

class GitUnknownBranchNameError(Exception):
    pass


class GitRepository(Repository):
    """A class representing a GIT repository.

    Note that this is just the parent directory for a bunch of git
    branches, making it easy to switch to a mirror URI.
    """

    init_xml_attrs = ['href']

    def __init__(self, config, name, href):
        Repository.__init__(self, config, name)
        # allow user to adjust location of branch.
        self.href = config.repos.get(name, href)

    branch_xml_attrs = ['module', 'subdir', 'checkoutdir', 'revision', 'tag']

    def branch(self, name, module = None, subdir="", checkoutdir = None,
               revision = None, tag = None):
        if module is None:
            module = name

        mirror_module = None
        if self.config.dvcs_mirror_dir:
            mirror_module = get_git_mirror_directory(
                    self.config.dvcs_mirror_dir, checkoutdir, module)

        # allow remapping of branch for module, it supports two modes of
        # operation
        if name in self.config.branches:
            branch_mapping = self.config.branches.get(name)
            if type(branch_mapping) is str:
                # passing a single string will override the branch name
                revision = branch_mapping
            else:
                # otherwise it is assumed it is a pair, redefining both
                # git URI and the branch to use
                try:
                    new_module, revision = self.config.branches.get(name)
                except (ValueError, TypeError):
                    logging.warning(_('ignored bad branch redefinition for module:') + ' ' + name)
                else:
                    if new_module:
                        module = new_module
        if not (urlparse.urlparse(module)[0] or module[0] == '/'):
            if self.href.endswith('/'):
                base_href = self.href
            else:
                base_href = self.href + '/'
            module = base_href + module

        if mirror_module:
            return GitBranch(self, mirror_module, subdir, checkoutdir,
                    revision, tag, unmirrored_module=module)
        else:
            return GitBranch(self, module, subdir, checkoutdir, revision, tag)

    def to_sxml(self):
        return [sxml.repository(type='git', name=self.name, href=self.href)]

    def get_sysdeps(self):
        return ['git']


class GitBranch(Branch):
    """A class representing a GIT branch."""

    dirty_branch_suffix = '-dirty'

    def __init__(self, repository, module, subdir, checkoutdir=None,
                 branch=None, tag=None, unmirrored_module=None):
        Branch.__init__(self, repository, module, checkoutdir)
        self.subdir = subdir
        self.branch = branch
        self.tag = tag
        self.unmirrored_module = unmirrored_module

    def get_module_basename(self):
        # prevent basename() from returning empty strings on trailing '/'
        name = self.module.rstrip(os.sep)
        name = os.path.basename(name)
        if name.endswith('.git'):
            name = name[:-4]
        return name
 
    def srcdir(self):
        path_elements = [self.checkoutroot]
        if self.checkoutdir:
            path_elements.append(self.checkoutdir)
        else:
            path_elements.append(self.get_module_basename())
        if self.subdir:
            path_elements.append(self.subdir)
        return os.path.join(*path_elements)
    srcdir = property(srcdir)

    def branchname(self):
        return self.branch
    branchname = property(branchname)

    def execute_git_predicate(self, predicate):
        """A git command wrapper for the cases, where only the boolean outcome
        is of interest.
        """
        try:
            get_output(predicate, cwd=self.get_checkoutdir(),
                    extra_env=get_git_extra_env())
        except CommandError:
            return False
        return True

    def is_local_branch(self, branch):
        is_local_head = self.execute_git_predicate( ['git', 'show-ref', '--quiet',
                                                     '--verify', 'refs/heads/' + branch])
        if is_local_head:
            return True
        return self.execute_git_predicate(['git', 'rev-parse', branch])

    def is_inside_work_tree(self):
        return self.execute_git_predicate(
                ['git', 'rev-parse', '--is-inside-work-tree'])

    def is_tracking_a_remote_branch(self, local_branch):
        if not local_branch:
            return False
        current_branch_remote_config = 'branch.%s.remote' % local_branch
        return self.execute_git_predicate(
                ['git', 'config', '--get', current_branch_remote_config])

    def is_dirty(self, ignore_submodules=True):
        submodule_options = []
        if ignore_submodules:
            if not self.check_version_git('1.5.6'):
                raise CommandError(_('Need at least git-1.5.6 from June/08 '
                        'to operate'))
            submodule_options = ['--ignore-submodules']
        return not self.execute_git_predicate(
                ['git', 'diff', '--exit-code', '--quiet'] + submodule_options
                + ['HEAD'])

    def check_version_git(self, version_spec):
        return check_version(['git', '--version'], r'git version ([\d.]+)',
                version_spec, extra_env=get_git_extra_env())

    def get_current_branch(self):
        """Returns either a branchname or None if head is detached"""
        if not self.is_inside_work_tree():
            raise CommandError(_('Unexpected: Checkoutdir is not a git '
                    'repository:' + self.get_checkoutdir()))
        try:
            full_branch = get_output(['git', 'symbolic-ref', '-q', 'HEAD'],
                            cwd=self.get_checkoutdir(),
                            extra_env=get_git_extra_env()).strip()
            # strip refs/heads/ to get the branch name only
            return full_branch.replace('refs/heads/', '')
        except CommandError:
            return None

    def find_remote_branch_online_if_necessary(self, buildscript,
            remote_name, branch_name):
        """Try to find the given branch first, locally, then remotely, and state
        the availability in the return value."""
        wanted_ref = remote_name + '/' + branch_name
        if self.execute_git_predicate( ['git', 'show-ref', wanted_ref]):
            return True
        buildscript.execute(['git', 'fetch'], cwd=self.get_checkoutdir(),
                extra_env=get_git_extra_env())
        return self.execute_git_predicate( ['git', 'show-ref', wanted_ref])

    def get_branch_switch_destination(self):
        current_branch = self.get_current_branch()
        wanted_branch = self.branch or 'master'

        # Always switch away from a detached head.
        if not current_branch:
            return wanted_branch

        assert(current_branch and wanted_branch)
        # If the current branch is not tracking a remote branch it is assumed to
        # be a local work branch, and it won't be considered for a change.
        if current_branch != wanted_branch \
                and self.is_tracking_a_remote_branch(current_branch):
            return wanted_branch

        return None

    def switch_branch_if_necessary(self, buildscript):
        """
        The switch depends on the requested tag, the requested branch, and the
        state and type of the current branch.

        An imminent branch switch generates an error if there are uncommited
        changes.
        """
        wanted_branch = self.get_branch_switch_destination()
        switch_command = []
        if self.tag:
            switch_command= ['git', 'checkout', self.tag]
        elif wanted_branch:
            if self.is_local_branch(wanted_branch):
                switch_command = ['git', 'checkout', wanted_branch]
            else:
                if not self.find_remote_branch_online_if_necessary(
                        buildscript, 'origin', wanted_branch):
                    raise CommandError(_('The requested branch "%s" is '
                            'not available. Neither locally, nor remotely '
                            'in the origin remote.' % wanted_branch))
                switch_command = ['git', 'checkout', '--track', '-b',
                        wanted_branch, 'origin/' + wanted_branch]

        if switch_command:
            if self.is_dirty():
                raise CommandError(_('Refusing to switch a dirty tree.'))
            buildscript.execute(switch_command, cwd=self.get_checkoutdir(),
                    extra_env=get_git_extra_env())

    def rebase_current_branch(self, buildscript):
        """Pull the current branch if it is tracking a remote branch."""
        branch = self.get_current_branch();
        if not self.is_tracking_a_remote_branch(branch):
            return

        git_extra_args = {'cwd': self.get_checkoutdir(),
                'extra_env': get_git_extra_env()}

        stashed = False
        if self.is_dirty(ignore_submodules=True):
            stashed = True
            buildscript.execute(['git', 'stash', 'save', 'jhbuild-stash'],
                    **git_extra_args)

        buildscript.execute(['git', 'rebase', 'origin/' + branch],
                            **git_extra_args)

        if stashed:
            # git stash pop was introduced in 1.5.5,
            if self.check_version_git('1.5.5'):
                buildscript.execute(['git', 'stash', 'pop'], **git_extra_args)
            else:
                buildscript.execute(['git', 'stash', 'apply', 'jhbuild-stash'],
                        **git_extra_args)

    def move_to_sticky_date(self, buildscript):
        if self.config.quiet_mode:
            quiet = ['-q']
        else:
            quiet = []
        commit = self._get_commit_from_date()
        branch = 'jhbuild-date-branch'
        branch_cmd = ['git', 'checkout'] + quiet + [branch]
        git_extra_args = {'cwd': self.get_checkoutdir(),
                'extra_env': get_git_extra_env()}
        if self.config.sticky_date == 'none':
            current_branch = self.get_current_branch()
            if current_branch and current_branch == branch:
                buildscript.execute(['git', 'checkout'] + quiet + ['master'],
                        **git_extra_args)
            return
        try:
            buildscript.execute(branch_cmd, **git_extra_args)
        except CommandError:
            branch_cmd = ['git', 'checkout'] + quiet + ['-b', branch]
            buildscript.execute(branch_cmd, **git_extra_args)
        buildscript.execute(['git', 'reset', '--hard', commit], **git_extra_args)

    def get_remote_branches_list(self):
        return [x.strip() for x in get_output(['git', 'branch', '-r'],
                cwd=self.get_checkoutdir(),
                extra_env=get_git_extra_env()).splitlines()]

    def exists(self):
        try:
            refs = get_output(['git', 'ls-remote', self.module],
                    extra_env=get_git_extra_env())
        except:
            return False

        #FIXME: Parse output from ls-remote to work out if tag/branch is present

        return True

    def _get_commit_from_date(self):
        cmd = ['git', 'log', '--max-count=1', '--first-parent',
                '--until=%s' % self.config.sticky_date, 'master']
        cmd_desc = ' '.join(cmd)
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                                cwd=self.get_checkoutdir(),
                                env=get_git_extra_env())
        stdout = proc.communicate()[0]
        if not stdout.strip():
            raise CommandError(_('Command %s returned no output') % cmd_desc)
        for line in stdout.splitlines():
            if line.startswith('commit '):
                commit = line.split(None, 1)[1].strip()
                return commit
        raise CommandError(_('Command %s did not include commit line: %r')
                           % (cmd_desc, stdout))

    def _export(self, buildscript):
        self._checkout(buildscript)

        try:
            output = get_output(['git', 'rev-parse', 'HEAD'],
                    cwd = self.get_checkoutdir(), get_stderr=False,
                    extra_env=get_git_extra_env())
            tag = output.strip()
        except:
            tag = 'unknown'

        filename = self.get_module_basename() + '-' + tag + '.zip'

        if self.config.export_dir is not None:
            path = os.path.join(self.config.export_dir, filename)
        else:
            path = os.path.join(self.checkoutroot, filename)

        git_extra_args = {'cwd': self.get_checkoutdir(), 'extra_env': get_git_extra_env()}
        buildscript.execute(['git', 'archive', '-o', path, 'HEAD'], **git_extra_args)

    def _update_submodules(self, buildscript):
        if os.path.exists(os.path.join(self.get_checkoutdir(), '.gitmodules')):
            cmd = ['git', 'submodule', 'init']
            buildscript.execute(cmd, cwd=self.get_checkoutdir(),
                    extra_env=get_git_extra_env())
            cmd = ['git', 'submodule', 'update']
            buildscript.execute(cmd, cwd=self.get_checkoutdir(),
                    extra_env=get_git_extra_env())

    def update_dvcs_mirror(self, buildscript):
        if not self.config.dvcs_mirror_dir:
            return
        if self.config.nonetwork:
            return

        # Calculate a new in case a configuration reload changed the mirror root.
        mirror_dir = get_git_mirror_directory(self.config.dvcs_mirror_dir,
                self.checkoutdir, self.unmirrored_module)

        if os.path.exists(mirror_dir):
            buildscript.execute(['git', 'remote', 'set-url', 'origin',
                    self.unmirrored_module], cwd=mirror_dir,
                    extra_env=get_git_extra_env())
            buildscript.execute(['git', 'fetch'], cwd=mirror_dir,
                    extra_env=get_git_extra_env())
        else:
            buildscript.execute(
                    ['git', 'clone', '--mirror', self.unmirrored_module,
                    mirror_dir], extra_env=get_git_extra_env())

    def _checkout(self, buildscript, copydir=None):

        extra_opts = []
        if self.config.quiet_mode:
            extra_opts.append('-q')

        if self.config.shallow_clone:
            extra_opts.append('--depth=1')

        self.update_dvcs_mirror(buildscript)

        cmd = ['git', 'clone'] + extra_opts + [self.module]
        if self.checkoutdir:
            cmd.append(self.checkoutdir)

        if self.branch is not None:
            cmd.extend(['-b', self.branch])

        if copydir:
            buildscript.execute(cmd, cwd=copydir, extra_env=get_git_extra_env())
        else:
            buildscript.execute(cmd, cwd=self.config.checkoutroot,
                    extra_env=get_git_extra_env())

        self._update(buildscript, copydir=copydir, update_mirror=False)


    def _update(self, buildscript, copydir=None, update_mirror=True):
        cwd = self.get_checkoutdir()
        git_extra_args = {'cwd': cwd, 'extra_env': get_git_extra_env()}

        if not os.path.exists(os.path.join(cwd, '.git')):
            if os.path.exists(os.path.join(cwd, '.svn')):
                raise CommandError(_('Failed to update module as it switched to git (you should check for changes then remove the directory).'))
            raise CommandError(_('Failed to update module (missing .git) (you should check for changes then remove the directory).'))

        if update_mirror:
            self.update_dvcs_mirror(buildscript)

        buildscript.execute(['git', 'remote', 'set-url', 'origin',
                self.module], **git_extra_args)

        buildscript.execute(['git', 'remote', 'update', 'origin'],
                **git_extra_args)

        if self.config.sticky_date:
            self.move_to_sticky_date(buildscript)

        self.switch_branch_if_necessary(buildscript)

        self.rebase_current_branch(buildscript)

        self._update_submodules(buildscript)

    def may_checkout(self, buildscript):
        if buildscript.config.nonetwork and not buildscript.config.dvcs_mirror_dir:
            return False
        return True

    def checkout(self, buildscript):
        if not inpath('git', os.environ['PATH'].split(os.pathsep)):
            raise CommandError(_('%s not found') % 'git')
        Branch.checkout(self, buildscript)

    def delete_unknown_files(self, buildscript):
        git_extra_args = {'cwd': self.get_checkoutdir(), 'extra_env': get_git_extra_env()}
        buildscript.execute(['git', 'clean', '-d', '-f', '-x'], **git_extra_args)

    def tree_id(self):
        if not os.path.exists(self.get_checkoutdir()):
            return None
        try:
            output = get_output(['git', 'rev-parse', 'HEAD'],
                    cwd = self.get_checkoutdir(), get_stderr=False,
                    extra_env=get_git_extra_env())
        except CommandError:
            return None
        except GitUnknownBranchNameError:
            return None
        id_suffix = ''
        if self.is_dirty():
            id_suffix = self.dirty_branch_suffix
        return output.strip() + id_suffix

    def to_sxml(self):
        attrs = {}
        if self.branch:
            attrs['revision'] = self.branch
        if self.checkoutdir:
            attrs['checkoutdir'] = self.checkoutdir
        if self.subdir:
            attrs['subdir'] = self.subdir
        return [sxml.branch(repo=self.repository.name,
                            module=self.module,
                            tag=self.tree_id(),
                            **attrs)]


class GitSvnBranch(GitBranch):
    def __init__(self, repository, module, checkoutdir, revision=None):
        GitBranch.__init__(self, repository, module, "", checkoutdir, branch="git-svn")
        self.revision = revision

    def may_checkout(self, buildscript):
        return Branch.may_checkout(self, buildscript)

    def _get_externals(self, buildscript, branch="git-svn"):
        cwd = self.get_checkoutdir()
        try:
            externals = {}
            match = None
            external_expr = re.compile(r"\+dir_prop: (.*?) svn:externals (.*)$")
            rev_expr = re.compile(r"^r(\d+)$")
            # the unhandled.log file has the revision numbers
            # encoded as r#num we should only parse as far as self.revision
            for line in open(os.path.join(cwd, '.git', 'svn', branch, 'unhandled.log')):
                m = external_expr.search(line)
                if m:
                    match = m
                rev_match = rev_expr.search(line)
                if self.revision and rev_match:
                    if rev_match.group(1) > self.revision:
                        break
        except IOError:
            # we couldn't find an unhandled.log to parse so try
            # git svn show-externals - note this is broken in git < 1.5.6
            try:
                output = get_output(['git', 'svn', 'show-externals'], cwd=cwd,
                        extra_env=get_git_extra_env())
                # we search for comment lines to strip them out
                comment_line = re.compile(r"^#.*")
                ext = ''
                for line in output.splitlines():
                    if not comment_line.search(line):
                        ext += ' ' + line

                match = re.compile("^(\.) (.+)").search(". " + ext)
            except OSError:
                raise FatalError(_("External handling failed\n If you are running git version < 1.5.6 it is recommended you update.\n"))

        # only parse the final match
        if match:
            branch = match.group(1)
            external = urllib.unquote(match.group(2).replace("%0A", " ").strip("%20 ")).split()
            revision_expr = re.compile(r"-r(\d*)")
            i = 0
            while i < len(external):
                # see if we have a revision number
                match = revision_expr.search(external[i+1])
                if match:
                    externals[external[i]] = (external[i+2], match.group(1))
                    i = i+3
                else:
                    externals[external[i]] = (external[i+1], None)
                    i = i+2
        
        for extdir in externals.iterkeys():
            uri = externals[extdir][0]
            revision = externals[extdir][1]
            extdir = cwd+os.sep+extdir
            # FIXME: the "right way" is to use submodules
            extbranch = GitSvnBranch(self.repository, uri, extdir, revision)
            
            try:
                os.stat(extdir)[stat.ST_MODE]
                extbranch._update(buildscript)
            except OSError:
                extbranch._checkout(buildscript)

    def _checkout(self, buildscript, copydir=None):
        if self.config.sticky_date:
            raise FatalError(_('date based checkout not yet supported\n'))

        cmd = ['git', 'svn', 'clone', self.module]
        if self.checkoutdir:
            cmd.append(self.checkoutdir)

        # FIXME (add self.revision support)
        try:
            last_revision = jhbuild.versioncontrol.svn.get_info (self.module)['last changed rev']
            if not self.revision:
                cmd.extend(['-r', last_revision])
        except KeyError:
            raise FatalError(_('Cannot get last revision from %s. Check the module location.') % self.module)

        if copydir:
            buildscript.execute(cmd, cwd=copydir,
                    extra_env=get_git_extra_env())
        else:
            buildscript.execute(cmd, cwd=self.config.checkoutroot,
                    extra_env=get_git_extra_env())

        try:
            # is known to fail on some versions
            cmd = ['git', 'svn', 'show-ignore']
            s = get_output(cmd, cwd = self.get_checkoutdir(copydir),
                    extra_env=get_git_extra_env())
            fd = file(os.path.join(
                        self.get_checkoutdir(copydir), '.git/info/exclude'), 'a')
            fd.write(s)
            fd.close()
            buildscript.execute(cmd, cwd=self.get_checkoutdir(copydir),
                    extra_env=get_git_extra_env())
        except:
            pass

        # FIXME, git-svn should support externals
        self._get_externals(buildscript, self.branch)

    def _update(self, buildscript, copydir=None):
        if self.config.sticky_date:
            raise FatalError(_('date based checkout not yet supported\n'))

        if self.config.quiet_mode:
            quiet = ['-q']
        else:
            quiet = []

        cwd = self.get_checkoutdir()
        git_extra_args = {'cwd': cwd, 'extra_env': get_git_extra_env()}

        last_revision = get_output(['git', 'svn', 'find-rev', 'HEAD'],
                **git_extra_args)

        stashed = False
        if get_output(['git', 'diff'], **git_extra_args):
            # stash uncommitted changes on the current branch
            stashed = True
            buildscript.execute(['git', 'stash', 'save', 'jhbuild-stash'],
                    **git_extra_args)

        buildscript.execute(['git', 'checkout'] + quiet + ['master'],
                **git_extra_args)
        buildscript.execute(['git', 'svn', 'rebase'], **git_extra_args)

        if stashed:
            buildscript.execute(['git', 'stash', 'pop'], **git_extra_args)

        current_revision = get_output(['git', 'svn', 'find-rev', 'HEAD'],
                **git_extra_args)

        if last_revision != current_revision:
            try:
                # is known to fail on some versions
                cmd = "git svn show-ignore >> .git/info/exclude"
                buildscript.execute(cmd, **git_extra_args)
            except:
                pass

        # FIXME, git-svn should support externals
        self._get_externals(buildscript, self.branch)

class GitCvsBranch(GitBranch):
    def __init__(self, repository, module, checkoutdir, revision=None):
        GitBranch.__init__(self, repository, module, "", checkoutdir)
        self.revision = revision

    def may_checkout(self, buildscript):
        return Branch.may_checkout(self, buildscript)

    def branchname(self):
        for b in ['remotes/' + str(self.branch), self.branch, 'trunk', 'master']:
            if self.branch_exist(b):
                return b
        raise
    branchname = property(branchname)

    def _checkout(self, buildscript, copydir=None):

        cmd = ['git', 'cvsimport', '-r', 'cvs', '-p', 'b,HEAD',
               '-k', '-m', '-a', '-v', '-d', self.repository.cvsroot, '-C']

        if self.checkoutdir:
            cmd.append(self.checkoutdir)
        else:
            cmd.append(self.module)

        cmd.append(self.module)

        if copydir:
            buildscript.execute(cmd, cwd=copydir, extra_env=get_git_extra_env())
        else:
            buildscript.execute(cmd, cwd=self.config.checkoutroot,
                    extra_env=get_git_extra_env())

    def _update(self, buildscript, copydir=None):
        if self.config.sticky_date:
            raise FatalError(_('date based checkout not yet supported\n'))

        cwd = self.get_checkoutdir()
        git_extra_args = {'cwd': cwd, 'extra_env': get_git_extra_env()}

        stashed = False
        # stash uncommitted changes on the current branch
        if get_output(['git', 'diff'], **git_extra_args):
            # stash uncommitted changes on the current branch
            stashed = True
            buildscript.execute(['git', 'stash', 'save', 'jhbuild-stash'],
                    **git_extra_args)

        self._checkout(buildscript, copydir=copydir)

        if stashed:
            buildscript.execute(['git', 'stash', 'pop'], **git_extra_args)

register_repo_type('git', GitRepository)



# jhbuild - a tool to ease building collections of source packages
# Copyright (C) 2001-2006  James Henstridge
#
#   perl.py: perl module type definitions.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

__metaclass__ = type

import os

from jhbuild.errors import BuildStateError
from jhbuild.modtypes import \
     Package, DownloadableModule, register_module_type

__all__ = [ 'PerlModule' ]

class PerlModule(Package, DownloadableModule):
    """Base type for modules that are distributed with a Perl style
    "Makefile.PL" Makefile."""
    type = 'perl'

    PHASE_CHECKOUT = DownloadableModule.PHASE_CHECKOUT
    PHASE_FORCE_CHECKOUT = DownloadableModule.PHASE_FORCE_CHECKOUT
    PHASE_BUILD = 'build'
    PHASE_INSTALL = 'install'

    def __init__(self, name, branch=None, makeargs=''):
        Package.__init__(self, name, branch=branch)
        self.makeargs = makeargs

    def get_srcdir(self, buildscript):
        return self.branch.srcdir

    def get_builddir(self, buildscript):
        # does not support non-srcdir builds
        return self.get_srcdir(buildscript)

    def do_build(self, buildscript):
        buildscript.set_action(_('Building'), self)
        builddir = self.get_builddir(buildscript)
        perl = os.environ.get('PERL', 'perl')
        make = os.environ.get('MAKE', 'make')
        makeargs = self.makeargs + ' ' + self.config.module_makeargs.get(
                self.name, self.config.makeargs)
        cmd = '%s Makefile.PL INSTALLDIRS=vendor PREFIX=%s %s' % (perl, buildscript.config.prefix, makeargs)
        buildscript.execute(cmd, cwd=builddir, extra_env = self.extra_env)
        buildscript.execute([make, 'LD_RUN_PATH='], cwd=builddir,
                extra_env = self.extra_env)
    do_build.depends = [PHASE_CHECKOUT]
    do_build.error_phases = [PHASE_FORCE_CHECKOUT]

    def do_install(self, buildscript):
        buildscript.set_action(_('Installing'), self)
        builddir = self.get_builddir(buildscript)
        make = os.environ.get('MAKE', 'make')
        buildscript.execute(
                [make, 'install', 'PREFIX=%s' % buildscript.config.prefix],
                cwd = builddir, extra_env = self.extra_env)
        buildscript.moduleset.packagedb.add(self.name,
                                            self.get_revision() or '',
                                            self.get_destdir(buildscript))
    do_install.depends = [PHASE_BUILD]

    def xml_tag_and_attrs(self):
        return 'perl', [('id', 'name', None),
                         ('makeargs', 'makeargs', '')]


def parse_perl(node, config, uri, repositories, default_repo):
    instance = PerlModule.parse_from_xml(node, config, uri, repositories, default_repo)

    if node.hasAttribute('makeargs'):
        makeargs = node.getAttribute('makeargs')
        instance.makeargs = instance.eval_args(makeargs)

    return instance
register_module_type('perl', parse_perl)




# jhbuild - a tool to ease building collections of source packages
# Copyright (C) 2001-2006  James Henstridge
#
#   packagedb.py - a registry of installed packages
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import os
import sys
import time
import logging
import errno
import xml.dom.minidom as DOM
try:
    import hashlib
except ImportError:
    import md5 as hashlib

try:
    import xml.etree.ElementTree as ET
except ImportError:
    import elementtree.ElementTree as ET

from StringIO import StringIO

from jhbuild.utils import fileutils

def _parse_isotime(string):
    if string[-1] != 'Z':
        return time.mktime(time.strptime(string, '%Y-%m-%dT%H:%M:%S'))
    tm = time.strptime(string, '%Y-%m-%dT%H:%M:%SZ')
    return time.mktime(tm[:8] + (0,)) - time.timezone    

def _format_isotime(tm):
    return time.strftime('%Y-%m-%dT%H:%M:%SZ', time.gmtime(tm))

class PackageEntry:
    def __init__(self, package, version, metadata, dirname):
        self.package = package # string
        self.version = version # string
        self.metadata = metadata # hash of string to value
        self.dirname = dirname

    _manifest = None
    def get_manifest(self):
        if self._manifest:
            return self._manifest
        if not os.path.exists(os.path.join(self.dirname, 'manifests', self.package)):
            return None
        self._manifest = []
        for line in file(os.path.join(self.dirname, 'manifests', self.package)):
            self._manifest.append(line.strip())
        return self._manifest

    def set_manifest(self, value):
        if value is None:
            self._manifest = value
            return
        self._manifest = [x.strip() for x in value if not '\n' in value]
        if len(self._manifest) != len(value):
            logging.error(_('package %s has files with embedded new lines') % self.package)

    manifest = property(get_manifest, set_manifest)

    def write(self):
        # write info file
        fileutils.mkdir_with_parents(os.path.join(self.dirname, 'info'))
        writer = fileutils.SafeWriter(os.path.join(self.dirname, 'info', self.package))
        ET.ElementTree(self.to_xml()).write(writer.fp)
        writer.fp.write('\n')
        writer.commit()

        # write manifest
        fileutils.mkdir_with_parents(os.path.join(self.dirname, 'manifests'))
        writer = fileutils.SafeWriter(os.path.join(self.dirname, 'manifests', self.package))
        writer.fp.write('\n'.join(self.manifest).encode('utf-8', 'backslashreplace') + '\n')
        writer.commit()

    def remove(self):
        # remove info file
        fileutils.ensure_unlinked(os.path.join(self.dirname, 'info', self.package))

        # remove manifest
        fileutils.ensure_unlinked(os.path.join(self.dirname, 'manifests', self.package))

    def to_xml(self):
        entry_node = ET.Element('entry', {'package': self.package,
                                          'version': self.version})
        if 'installed-date' in self.metadata:
            entry_node.attrib['installed'] = _format_isotime(self.metadata['installed-date'])
        if 'configure-hash' in self.metadata:
            entry_node.attrib['configure-hash'] = \
                self.metadata['configure-hash']

        return entry_node

    @classmethod
    def from_xml(cls, node, dirname):
        package = node.attrib['package']
        version = node.attrib['version']
        metadata = {}

        installed_string = node.attrib['installed']
        if installed_string:
            metadata['installed-date'] = _parse_isotime(installed_string)
        configure_hash = node.attrib.get('configure-hash')
        if configure_hash:
            metadata['configure-hash'] = configure_hash

        dbentry = cls(package, version, metadata, dirname)

        return dbentry

    @classmethod
    def open(cls, dirname, package):
        try:
            info = open (os.path.join (dirname, 'info', package))
            doc = ET.parse(info)
            node = doc.getroot()

            if node.tag == 'entry':
                return PackageEntry.from_xml(node, dirname)

        except EnvironmentError as e:
            if e.errno != errno.ENOENT:
                raise

        # That didn't work: try the old packagedb.xml file instead.  We
        # use the manifest file to check if the package 'really exists'
        # because otherwise we may see the old packagedb.xml entry for
        # an uninstalled package (since we no longer update that file)
        #
        # please delete this code in 2016
        try:
            if os.path.exists(os.path.join(dirname, 'manifests', package)):
                info = open (os.path.join (dirname, 'packagedb.xml'))
                doc = ET.parse(info)
                root = doc.getroot()

                if root.tag == 'packagedb':
                    for node in root:
                        if node.tag == 'entry' and node.attrib['package'] == package:
                            return PackageEntry.from_xml(node, dirname)

        except EnvironmentError as e:
            if e.errno != errno.ENOENT:
                raise

        # it seems not to exist...
        return None

class PackageDB:
    def __init__(self, dbfile, config):
        self.dirname = os.path.dirname(dbfile)
        self.config = config

    def get(self, package):
        '''Return entry if package is installed, otherwise return None.'''
        return PackageEntry.open(self.dirname, package)

    def add(self, package, version, contents, configure_cmd = None):
        '''Add a module to the install cache.'''
        entry = self.get(package)
        if entry:
            metadata = entry.metadata
        else:
            metadata = {}
        metadata['installed-date'] = time.time() # now
        if configure_cmd:
            metadata['configure-hash'] = hashlib.md5(configure_cmd).hexdigest()
        pkg = PackageEntry(package, version, metadata, self.dirname)
        pkg.manifest = contents
        pkg.write()

    def check(self, package, version=None):
        '''Check whether a particular module is installed.'''
        entry = self.get(package)
        if entry is None:
            return False
        if version is not None:
            if entry.version != version: return False
        return True

    def installdate(self, package, version=None):
        '''Get the install date for a particular module.'''
        entry = self.get(package)
        if entry is None:
            return None
        if version and (entry.version != version):
            return None
        return entry.metadata['installed-date']

    def uninstall(self, package_name):
        '''Remove a module from the install cache.'''
        entry = self.get(package_name)

        if entry is None:
            raise KeyError

        if entry.manifest is None:
            logging.error(_("no manifest for '%s', can't uninstall.  Try building again, then uninstalling.") % (package_name,))
            return

        # Skip files that aren't in the prefix; otherwise we
        # may try to remove the user's ~ or something
        # (presumably we'd fail, but better not to try)
        to_delete = fileutils.filter_files_by_prefix(self.config, entry.manifest)

        # Don't warn on non-empty directories; we want to allow multiple
        # modules to share the same directory.  We could improve this by
        # reference-counting directories.
        for (path, was_deleted, error_string) in fileutils.remove_files_and_dirs(to_delete, allow_nonempty_dirs=True):
            if was_deleted:
                logging.info(_("Deleted: %(file)r") % {'file': path})
            elif error_string is None:
                pass
            else:
                logging.warn(_("Failed to delete %(file)r: %(msg)s") % { 'file': path,
                                                                         'msg': error_string})

        entry.remove()



# jhbuild - a tool to ease building collections of source packages
# Copyright (C) 2001-2006  James Henstridge
# Copyright (C) 2003-2004  Seth Nickell
#
#   terminal.py: build logic for a terminal interface
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import sys
import os
import signal
import subprocess
import unicodedata
import locale

from jhbuild.frontends import buildscript
from jhbuild.utils import cmds
from jhbuild.utils import trayicon
from jhbuild.utils import notify
from jhbuild.errors import CommandError, FatalError

term = os.environ.get('TERM', '')
is_xterm = term.find('xterm') >= 0 or term == 'rxvt'
del term

try: t_bold = cmds.get_output(['tput', 'bold'])
except:
    try: t_bold = cmds.get_output(['tput', 'md'])
    except: t_bold = ''

try: t_reset = cmds.get_output(['tput', 'sgr0'])
except:
    try: t_reset = cmds.get_output(['tput', 'me'])
    except: t_reset = ''

t_colour = [''] * 16
try:
    for i in range(8):
        t_colour[i] = cmds.get_output(['tput', 'setf', '%d' % i])
        t_colour[i+8] = t_bold + t_colour[i]
except:
    try:
        for index, i in enumerate([0, 4, 2, 6, 1, 5, 3, 7]):
            t_colour[index] = cmds.get_output(['tput', 'setaf', '%d' % i])
            t_colour[index+8] = t_bold + t_colour[index]
    except:
        try:
            for index, i in enumerate([0, 4, 2, 6, 1, 5, 3, 7]):
                t_colour[index] = cmds.get_output(['tput', 'AF', '%d' % i])
                t_colour[index+8] = t_bold + t_colour[index]
        except:
            pass


user_shell = os.environ.get('SHELL', '/bin/sh')

try:
    import curses
except ImportError:
    curses = None
else:
    try:
        curses.setupterm()
    except:
        pass

# tray icon stuff ...
if DATADIR:
    icondir = os.path.join(DATADIR, 'jhbuild')
else:
    icondir = os.path.join(os.path.dirname(__file__), 'icons')
phase_map = {
    'checkout':       'checkout.png',
    'force_checkout': 'checkout.png',
    'download':       'checkout.png',
    'unpack':         'checkout.png',
    'patch':          'checkout.png',
    'configure':      'configure.png',
    #'clean':          'clean.png',
    'build':          'build.png',
    'check':          'check.png',
    'install':        'install.png',
    }

class TerminalBuildScript(buildscript.BuildScript):
    triedcheckout = None
    is_end_of_build = False

    def __init__(self, config, module_list, module_set=None):
        buildscript.BuildScript.__init__(self, config, module_list, module_set=module_set)
        self.trayicon = trayicon.TrayIcon(config)
        self.notify = notify.Notify(config)
        
    def message(self, msg, module_num=-1):
        '''Display a message to the user'''
        
        if module_num == -1:
            module_num = self.module_num
        if module_num > 0:
            progress = ' [%d/%d]' % (module_num, len(self.modulelist))
        else:
            progress = ''

        if not (self.config.quiet_mode and self.config.progress_bar):
            uprint('%s*** %s ***%s%s' % (t_bold, msg, progress, t_reset))
        else:
            progress_percent = 1.0 * (module_num-1) / len(self.modulelist)
            self.display_status_line(progress_percent, module_num, msg)

        if is_xterm:
            sys.stdout.write('\033]0;jhbuild:%s%s\007' % (uencode(msg), progress))
            sys.stdout.flush()
        self.trayicon.set_tooltip('%s%s' % (msg, progress))

    def set_action(self, action, module, module_num=-1, action_target=None):
        if module_num == -1:
            module_num = self.module_num
        if not action_target:
            action_target = module.name
        self.message('%s %s' % (action, action_target), module_num)

    def display_status_line(self, progress, module_num, message):
        if self.is_end_of_build:
            # hardcode progress to 100% at the end of the build
            progress = 1

        columns = curses.tigetnum('cols')
        width = columns / 2
        num_hashes = int(round(progress * width))
        progress_bar = '[' + (num_hashes * '=') + ((width - num_hashes) * '-') + ']'

        module_no_digits = len(str(len(self.modulelist)))
        format_str = '%%%dd' % module_no_digits
        module_pos = '[' + format_str % module_num + '/' + format_str % len(self.modulelist) + ']'

        output = '%s %s %s%s%s' % (progress_bar, module_pos, t_bold, message, t_reset)
        text_width = len(output) - (len(t_bold) + len(t_reset))
        if text_width > columns:
            output = output[:columns+len(t_bold)] + t_reset
        else:
            output += ' ' * (columns-text_width)

        sys.stdout.write('\r'+output)
        if self.is_end_of_build:
            sys.stdout.write('\n')
        sys.stdout.flush()


    def execute(self, command, hint=None, cwd=None, extra_env=None):
        if not command:
            raise CommandError(_('No command given'))

        kws = {
            'close_fds': True
            }
        print_args = {'cwd': ''}
        if cwd:
            print_args['cwd'] = cwd
        else:
            try:
                print_args['cwd'] = os.getcwd()
            except OSError:
                pass

        if isinstance(command, (str, unicode)):
            kws['shell'] = True
            print_args['command'] = command
        else:
            print_args['command'] = ' '.join(command)

        # get rid of hint if pretty printing is disabled.
        if not self.config.pretty_print:
            hint = None
        elif os.name == 'nt':
            # pretty print also doesn't work on Windows;
            # see https://bugzilla.gnome.org/show_bug.cgi?id=670349 
            hint = None

        if not self.config.quiet_mode:
            if self.config.print_command_pattern:
                try:
                    print self.config.print_command_pattern % print_args
                except TypeError as e:
                    raise FatalError('\'print_command_pattern\' %s' % e)
                except KeyError as e:
                    raise FatalError(_('%(configuration_variable)s invalid key'
                                       ' %(key)s' % \
                                       {'configuration_variable' :
                                            '\'print_command_pattern\'',
                                        'key' : e}))

        kws['stdin'] = subprocess.PIPE
        if hint in ('cvs', 'svn', 'hg-update.py'):
            kws['stdout'] = subprocess.PIPE
            kws['stderr'] = subprocess.STDOUT
        else:
            kws['stdout'] = None
            kws['stderr'] = None

        if self.config.quiet_mode:
            kws['stdout'] = subprocess.PIPE
            kws['stderr'] = subprocess.STDOUT

        if cwd is not None:
            kws['cwd'] = cwd

        if extra_env is not None:
            kws['env'] = os.environ.copy()
            kws['env'].update(extra_env)

        command = self._prepare_execute(command)

        try:
            p = subprocess.Popen(command, **kws)
        except OSError as e:
            raise CommandError(str(e))

        output = []
        if hint in ('cvs', 'svn', 'hg-update.py'):
            conflicts = []
            def format_line(line, error_output, conflicts = conflicts, output = output):
                if line.startswith('C '):
                    conflicts.append(line)

                if self.config.quiet_mode:
                    output.append(line)
                    return

                if line[-1] == '\n': line = line[:-1]

                if line.startswith('C '):
                    print '%s%s%s' % (t_colour[12], line, t_reset)
                elif line.startswith('M '):
                    print '%s%s%s' % (t_colour[10], line, t_reset)
                elif line.startswith('? '):
                    print '%s%s%s' % (t_colour[8], line, t_reset)
                else:
                    print line

            cmds.pprint_output(p, format_line)
            if conflicts:
                uprint(_('\nConflicts during checkout:\n'))
                for line in conflicts:
                    sys.stdout.write('%s  %s%s\n'
                                     % (t_colour[12], line, t_reset))
                # make sure conflicts fail
                if p.returncode == 0 and hint == 'cvs': p.returncode = 1
        elif self.config.quiet_mode:
            def format_line(line, error_output, output = output):
                output.append(line)
            cmds.pprint_output(p, format_line)
        else:
            try:
                p.communicate()
            except KeyboardInterrupt:
                try:
                    os.kill(p.pid, signal.SIGINT)
                except OSError:
                    # process might already be dead.
                    pass
        try:
            if p.wait() != 0:
                if self.config.quiet_mode:
                    print ''.join(output)
                raise CommandError(_('########## Error running %s')
                                   % print_args['command'], p.returncode)
        except OSError:
            # it could happen on a really badly-timed ctrl-c (see bug 551641)
            raise CommandError(_('########## Error running %s')
                               % print_args['command'])

    def start_module(self, module):
        self.triedcheckout = None

    def start_phase(self, module, phase):
        self.notify.clear()
        self.trayicon.set_icon(os.path.join(icondir,
                               phase_map.get(phase, 'build.png')))

    def end_build(self, failures):
        self.is_end_of_build = True
        if len(failures) == 0:
            self.message(_('success'))
        else:
            self.message(_('the following modules were not built'))
            for module in failures:
                print module,
            print

    def handle_error(self, module, phase, nextphase, error, altphases):
        '''handle error during build'''
        summary = _('Error during phase %(phase)s of %(module)s') % {
            'phase': phase, 'module':module.name}
        try:
            error_message = error.args[0]
            self.message('%s: %s' % (summary, error_message))
        except:
            error_message = None
            self.message(summary)
        self.trayicon.set_icon(os.path.join(icondir, 'error.png'))
        self.notify.notify(summary = summary, body = error_message,
                icon = 'dialog-error', expire = 20)

        if self.config.trycheckout:
            if self.triedcheckout is None and altphases.count('configure'):
                self.triedcheckout = 'configure'
                self.message(_('automatically retrying configure'))
                return 'configure'
            elif self.triedcheckout == 'configure' and altphases.count('force_checkout'):
                self.triedcheckout = 'done'
                self.message(_('automatically forcing a fresh checkout'))
                return 'force_checkout'
        self.triedcheckout = None

        if not self.config.interact:
            return 'fail'
        while True:
            print
            uprint('  [1] %s' % _('Rerun phase %s') % phase)
            if nextphase:
                uprint('  [2] %s' % _('Ignore error and continue to %s') % nextphase)
            else:
                uprint('  [2] %s' % _('Ignore error and continue to next module'))
            uprint('  [3] %s' % _('Give up on module'))
            uprint('  [4] %s' % _('Start shell'))
            uprint('  [5] %s' % _('Reload configuration'))
            nb_options = i = 6
            for altphase in altphases:
                try:
                    altphase_label = _(getattr(getattr(module, 'do_' + altphase), 'label'))
                except AttributeError:
                    altphase_label = altphase
                uprint('  [%d] %s' % (i, _('Go to phase "%s"') % altphase_label))
                i += 1
            val = raw_input(uencode(_('choice: ')))
            val = udecode(val)
            val = val.strip()
            if val == '1':
                return phase
            elif val == '2':
                return nextphase
            elif val == '3':
                return 'fail'
            elif val == '4':
                cwd = os.getcwd()
                builddir = module.get_builddir(self)
                srcdir = module.get_srcdir(self)
                try:
                    os.chdir(builddir)
                except OSError:
                    os.chdir(self.config.checkoutroot)
                unlink_srcdir = False
                try:
                    if builddir != srcdir:
                        os.symlink(srcdir, '.jhbuild-srcdir')
                        unlink_srcdir = True
                except OSError:
                    pass
                uprint(_('exit shell to continue with build'))
                os.system(user_shell)
                if unlink_srcdir:
                    try:
                        os.unlink('.jhbuild-srcdir')
                    except OSError:
                        pass
                os.chdir(cwd) # restor working directory
            elif val == '5':
                self.config.reload()
            else:
                try:
                    val = int(val)
                    selected_phase = altphases[val - nb_options]
                except:
                    uprint(_('invalid choice'))
                    continue
                try:
                    needs_confirmation = getattr(
                            getattr(module, 'do_' + selected_phase), 'needs_confirmation')
                except AttributeError:
                    needs_confirmation = False
                if needs_confirmation:
                    val = raw_input(uencode(_('Type "yes" to confirm the action: ')))
                    val = udecode(val)
                    val = val.strip()
                    def normalize(s):
                        return unicodedata.normalize('NFKD', s).encode('ascii', 'ignore').lower()
                    if normalize(val) in ('yes', normalize(_('yes')), _('yes').lower()):
                        return selected_phase
                    continue
                return selected_phase
        assert False, 'not reached'

BUILD_SCRIPT = TerminalBuildScript



# jhbuild - a tool to ease building collections of source packages
# Copyright Lieven Gobaerts
# Copyright (C)  2008 Igalia S.L., John Carr, Frederic Peters
#
#   feeds.py: RSS/Atom feeds
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# Minor changes made by API (apinheiro@igalia.com) in order to fit with our
# configuration and last buildbot changes

import time, re

from twisted.web.resource import Resource
from twisted.web import html as twhtml

from buildbot.status.builder import FAILURE, SUCCESS, WARNINGS

class XmlResource(Resource):
    contentType = "text/xml; charset=UTF-8"
    def render(self, request):
        data = self.content(request)
        request.setHeader("content-type", self.contentType)
        if request.method == "HEAD":
            request.setHeader("content-length", len(data))
            return ''
        return data
    docType = ''
    def header (self, request):
        data = ('<?xml version="1.0"?>\n')
        return data
    def footer(self, request):
        data = ''
        return data
    def content(self, request):
        data = self.docType
        data += self.header(request)
        data += self.body(request)
        data += self.footer(request)
        return data
    def body(self, request):
        return ''
    def getStatus(self, request):
        return request.site.buildbot_service.getStatus()
    def getControl(self, request):
        return request.site.buildbot_service.getControl()
    def getChangemaster(self, request):
        return request.site.buildbot_service.parent.change_svc

class FeedResource(XmlResource):
    title = 'Dummy (please reload)'
    link = 'http://dummylink'
    language = 'en'
    description = 'Dummy (please reload)'
    status = None

    def __init__(self, categories):
        self.categories = categories

    def getBuilds(self):
        builds = []
        builderNames = self.status.getBuilderNames(categories=self.categories)
        builders = map(lambda name: self.status.getBuilder(name), builderNames)
        maxFeeds = 5

        # Copy all failed builds in a new list.
        # This could clearly be implemented much better if we had access to a global list
        # of builds.
        for b in builders:
            lastbuild = b.getLastFinishedBuild()
            if lastbuild is None:
                continue

#           if b.category != "prod":
#                continue
            lastnr = lastbuild.getNumber()

            totalbuilds = 0
            i = lastnr
            while i >= 0:
                build = b.getBuild(i)
                i -= 1
                if not build:
                    continue

                results = build.getResults()

                # only add entries for failed builds!
                if results == FAILURE:
                    totalbuilds += 1
                    builds.append(build)

                # stop for this builder when our total nr. of feeds is reached
                if totalbuilds >= maxFeeds:
                    break

        # Sort build list by date, youngest first.
#        builds.sort(key=lambda build: build.getTimes(), reverse=True)

        # If you need compatibility with python < 2.4, use this for sorting instead:
          # We apply Decorate-Sort-Undecorate
        deco = [(build.getTimes(), build) for build in builds]
        deco.sort()
        deco.reverse()
        builds = [build for (b1, build) in deco]

        if builds:
            builds = builds[:min(len(builds), maxFeeds)]
        return builds

    def body (self, request):
        data = ''
        self.status = self.getStatus(request)
        self.link = str(self.status.getBuildbotURL())
        projectName = str(self.categories[0])
        self.title = 'Build status of %s' % projectName
        self.description = 'List of FAILed %s builds' % projectName

        builds = self.getBuilds()

        for build in builds:
            start, finished = build.getTimes()
            strFinished = time.strftime("%a, %d %b %Y %H:%M:%S %z", time.localtime(int(finished)))
            link = str(self.status.getURLForThing(build))

            # title: trunk r22191 (plus patch) failed on 'i686-debian-sarge1 shared gcc-3.3.5'
            ss = build.getSourceStamp()
            if ss is None:
                source = "[src unknown]"
            else:
                branch, revision, patch = ss.branch, ss.revision, ss.patch
                source = ""
                if branch:
                    source += "Branch %s " % branch
                if revision:
                    source += "r%s " % str(revision)
                else:
                    source += "trunk"
                if patch is not None:
                    source += " (plus patch)"
            builder_name = build.getBuilder().getName()[len(projectName)+1:]
            title = projectName + ' ' + source + " failed on '" + builder_name + "'"

            # get name of the failed step and the last 30 lines of its log.
            lastlog = ''
            laststep = None
            if build.getLogs():
                log = build.getLogs()[-1]
                laststep = log.getStep().getName()
                try:
                    lastlog = log.getText()
                except IOError:
                    # Probably the log file has been removed
                    lastlog='<b>log file not available</b>'

            lines = re.split('\n', lastlog)
            for logline in lines[max(0, len(lines)-30):]:
                lastlog = lastlog + logline

            description = '<dl>\n'
            description += '<dt>Date</dt><dd>%s</dd>\n' % strFinished
            description += '<dt>Build summary</dt><dd><a href="' + self.link + projectName + '">' + self.link + projectName + '</a></dd>\n'
            description += '<dt>Build details</dt><dd><a href="%s">%s</a></dd>' % (link, link)
            if build.getResponsibleUsers():
                description += '<dt>Author list</dt><dd>' + ', '.join(build.getResponsibleUsers()) + '</dd>\n'
            if laststep:
                description += '<dt>Failed step</dt><dd><b>%s</b></dd>\n' % laststep
            description += '</dl>\n'
            description += '<p>Last lines of the build log:</p>\n'
            description += '<pre>%s</pre>' % twhtml.escape(lastlog)

            data += self.item(title,
                              description = description,
                              link=link,pubDate=strFinished)

        if type(data) is unicode:
            data = data.encode('utf-8', 'ignore')
        return data

    def item(self, title='', link='', description='', pubDate=''):
        """Generates xml for one item in the feed"""

class Rss20StatusResource(FeedResource):
    def __init__(self, categories):
        FeedResource.__init__(self, categories)
        self.contentType = 'application/rss+xml'

    def header(self, request):
        data = FeedResource.header(self, request)
        data += '<rss version="2.0">\n'
        data += '<channel>'
        if self.title is not None:
            data += '<title>'+self.title+'</title>'
        if self.link is not None:
            data += '<link>'+self.link+'</link>'
        if self.language is not None:
            data += '<language>'+self.language+'</language>'
        if self.description is not None:
            data += '<description><![CDATA[%s]]></description>' % self.description
        return data

    def item(self, title='', link='', description='', pubDate=''):
        data = '<item>'
        data += '<title>'+title+'</title>'
        if link is not None:
            data += '<link>'+link+'</link>'
        if description is not None:
            data += '<description><![CDATA[%s]]></description>' % description
        if pubDate is not None:
            data += '<pubDate>'+pubDate+'</pubDate>'
        data += '</item>'
        return data

    def footer(self, request):
        data = ('</channel>'
                '</rss>')
        return data

class Atom10StatusResource(FeedResource):
    def __init__(self, categories):
        FeedResource.__init__(self, categories)
        self.contentType = 'application/atom+xml'

    def header(self, request):
        data = FeedResource.header(self, request)
        data += '<feed xmlns="http://www.w3.org/2005/Atom">\n'
        if self.title is not None:
            data += '<title>'+self.title+'</title>\n'
        if self.link is not None:
            data += '<link href="'+self.link+'"/>\n'
        # if self.language is not None:
            # data += '<language>'+self.language+'</language>'
        if self.description is not None:
            data += '<subtitle>'+self.description+'</subtitle>\n'
        return data

    def item(self, title='', link='', description='', pubDate=''):
        data = '<entry>\n'
        data += '<id>%s</id>\n' % link
        data += '<title>%s</title>\n' % title
        if link is not None:
            data += '<link href="%s"/>\n' % link
        if description is not None:
            data += '<content type="html">%s</content>\n' % twhtml.escape(description)
        if pubDate is not None:
            data += '<updated>%s</updated>\n' % pubDate
        data += '</entry>\n'
        return data

    def footer(self, request):
        data = ('</feed>')
        return data




# jhbuild - a tool to ease building collections of source packages
# Copyright (C) 2001-2006  James Henstridge
#
#   base.py: the most common jhbuild commands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from optparse import make_option

import jhbuild.moduleset
import jhbuild.frontends
from jhbuild.errors import UsageError, FatalError
from jhbuild.commands import Command, register_command


class cmd_clean(Command):
    doc = _('Clean all modules')

    name = 'clean'
    usage_args = '[ options ... ] [ modules ... ]'

    def __init__(self):
        Command.__init__(self, [
            make_option('-s', '--skip', metavar='MODULES',
                        action='append', dest='skip', default=[],
                        help=_('treat the given modules as up to date')),
            make_option('-t', '--start-at', metavar='MODULE',
                        action='store', dest='startat', default=None,
                        help=_('start building at the given module')),
            ])

    def run(self, config, options, args, help=None):
        for item in options.skip:
            config.skip += item.split(',')

        module_set = jhbuild.moduleset.load(config)
        module_list = module_set.get_module_list(args or config.modules, config.skip)
        # remove modules up to startat
        if options.startat:
            while module_list and module_list[0].name != options.startat:
                del module_list[0]
            if not module_list:
                raise FatalError(_('%s not in module list') % options.startat)

        # remove modules that are not marked as installed
        packagedb = module_set.packagedb
        for module in module_list[:]:
            if not packagedb.check(module.name):
                module_list.remove(module)

        config.nopoison = True

        build = jhbuild.frontends.get_buildscript(config, module_list, module_set=module_set)
        return build.build(phases=['clean'])

register_command(cmd_clean)
